__all__ = ['get_class_from_module_path']

from importlib import import_module


def get_class_from_module_path(path):
    try: module_path, class_name = path.rsplit('.', 1)
    except ValueError: raise ImportError('{} is not a valid module path. You need to specify the full Python dotted path to the module.'.format(path))

    module = import_module(module_path)
    klass = getattr(module, class_name)

    return klass
#end def



"""
This script is different from `ycml.scripts.partition_lines` because it takes labels into account and produces stratified partitions.
"""

from argparse import ArgumentParser
from collections import Counter
import json
import logging

from sklearn.model_selection import train_test_split

from uriutils import URIFileType

from ycml.utils import load_instances

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Script to partition instances in a stratified manner.')
    parser.add_argument('-i', '--instances', type=URIFileType('r'), nargs='+', metavar='<instances>', help='List of instance files to partition.')
    parser.add_argument('-l', '--label-key', type=str, metavar='<key>', default='labels', help='The key name for the label.')
    parser.add_argument('-s', '--train-size', type=float, default=0.7, metavar='<N>', help='Proportions of instances to use for training set.')
    parser.add_argument('-o', '--output', type=URIFileType('w'), nargs=2, required=True, metavar='<output>', help='Save partitioned instances here.')
    A = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s', level=logging.INFO)

    X, Y_labels = zip(*load_instances(A.instances, labels_field=A.label_key))

    labels_freq = Counter(label for labels in Y_labels for label in labels)

    if any(len(labels) > 1 for labels in Y_labels):
        logger.warning('These are multilabel instances, we will select the most common label for each instance to do stratification.')
        Y_single_labels = [max(labels, key=lambda l: labels_freq[l]) if labels else '<none>' for labels in Y_labels]
    else:
        Y_single_labels = [labels[0] if labels else '<none>' for labels in Y_labels]
    #end for

    labels_freq = Counter(Y_single_labels)
    removed_labels = []
    for i, label in enumerate(Y_single_labels):
        if labels_freq[label] < 2:
            removed_labels.append(label)
            Y_single_labels[i] = '<none>'
        #end if
    #end for
    removed_labels.sort()
    logger.warning('The following labels were replaced with `<none>` for the purpose of stratification: {}.'.format(', '.join(removed_labels)))

    X_train, X_test = train_test_split(X, train_size=A.train_size, stratify=Y_single_labels)

    for o in X_train:
        A.output[0].write(json.dumps(o))
        A.output[0].write('\n')
    #end for
    logger.info('{} training instances written to <{}>.'.format(len(X_train), A.output[0].name))

    for o in X_test:
        A.output[1].write(json.dumps(o))
        A.output[1].write('\n')
    #end for
    logger.info('{} evaluation instances written to <{}>.'.format(len(X_test), A.output[1].name))
#end def


if __name__ == '__main__': main()



from argparse import ArgumentParser
import json
import logging
import sys

from uriutils import URIFileType

from ..utils import load_dictionary_from_file, load_instances, get_settings
from ..featclass import load_featclass

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Predict instances using ML classifier.')
    parser.add_argument('-s', '--settings', type=URIFileType(), metavar='<settings_file>', help='Settings file to configure models.')
    parser.add_argument('instances', type=URIFileType('r'), metavar='<instances>', help='Instances to use for prediction.')
    parser.add_argument('--featclass', type=URIFileType(), metavar='<featclass_uri>', help='Featclass configuration file to use for prediction.')
    parser.add_argument('-p', '--probabilities', action='store_true', help='Also save prediction probabilities.')
    parser.add_argument('-o', '--output', type=URIFileType('w'), default=sys.stdout.buffer, help='Save predictions to this file.')

    A = parser.parse_args()

    file_settings = load_dictionary_from_file(A.settings) if A.settings else {}

    log_level = get_settings(key='log_level', sources=(A, 'env', file_settings), default='DEBUG').upper()
    log_format = get_settings(key='log_format', sources=(A, 'env', file_settings), default='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s')
    logging.basicConfig(format=log_format, level=logging.getLevelName(log_level))

    if A.featclass: featclass = load_featclass(settings=load_dictionary_from_file(A.featclass))
    else: featclass = load_featclass(settings=file_settings, uri=get_settings(key='featclass_uri', sources=('env', file_settings)))

    for count, args in enumerate(featclass.predictions_generator(load_instances(A.instances, labels_field=None), include_proba=A.probabilities, unbinarized=True), start=1):
        if A.probabilities:
            o, Y_predict_list, Y_proba_dict = args
            o['predictions'] = Y_predict_list
            o['probabilities'] = Y_proba_dict
        else:
            o, Y_predict_list = args
            o['predictions'] = Y_predict_list
        #end if

        A.output.write(json.dumps(o))
        A.output.write('\n')

        if count % 100000 == 0: logger.info('Saved {} predictions.'.format(count))
    #end for

    logger.info('Saved {} predictions to <{}>.'.format(count, A.output.name))
#end def


if __name__ == '__main__': main()



__all__ = ['CounterHashingVectorizer', 'ListHashingVectorizer', 'ListCountVectorizer', 'ListNGramAnalyzer', 'ListTfidfVectorizer', 'SpaceTokenizerTransformer', '_list_analyzer']

import logging

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import HashingVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer

from .base import PureTransformer

logger = logging.getLogger(__name__)


def _list_analyzer(L):
    for elem in L:
        yield elem
#end def


class ListNGramAnalyzer(PureTransformer):
    def __init__(self, ngram_range=(1, 1), ngram_delimiter=' ', **kwargs):
        kwargs.setdefault('nparray', False)
        super(ListNGramAnalyzer, self).__init__(**kwargs)

        if isinstance(ngram_range, tuple) and len(ngram_range) == 2: ngrams = list(range(ngram_range[0], ngram_range[1] + 1))
        elif isinstance(ngram_range, str): ngrams = [int(ngram_range)]
        elif isinstance(ngram_range, int): ngrams = [ngram_range]
        else: ngrams = ngram_range
        ngrams.sort()

        self.ngrams = ngrams
        self.ngram_delimiter = ngram_delimiter
    #end def

    def __call__(self, L):
        L = list(L)
        length = len(L)
        ngram_delimiter = self.ngram_delimiter

        for i in range(length):
            for n in self.ngrams:
                if i + n > length: break

                yield ngram_delimiter.join(L[i:i + n])
            #end for
        #end for
    #end def

    def transform_one(self, tokens, **kwargs):
        yield from self.__call__(tokens)
#end def


def _counter_analyzer(C):
    for k, v in C.items():
        for _ in range(v):
            yield k
#end def


class ListCountVectorizer(CountVectorizer):
    def __init__(self, **kwargs):
        kwargs.setdefault('analyzer', ListNGramAnalyzer(**kwargs))
        super(ListCountVectorizer, self).__init__(**kwargs)
    #end def

    def fit(self, *args, **kwargs):
        ret = super(ListCountVectorizer, self).fit(*args)
        logger.debug('There are {} vocabulary items in <{}>.'.format(len(self.vocabulary_), self))

        return ret
    #end def

    def fit_transform(self, *args, **kwargs):
        transformed = super(ListCountVectorizer, self).fit_transform(*args)
        logger.debug('There {} vocabulary items in <{}>.'.format(len(self.vocabulary_), self))

        return transformed
    #end def
#end class


class ListTfidfVectorizer(TfidfVectorizer):
    def __init__(self, **kwargs):
        kwargs.setdefault('analyzer', ListNGramAnalyzer(**kwargs))
        super(ListTfidfVectorizer, self).__init__(**kwargs)
    #end def

    def fit(self, *args, **kwargs):
        ret = super(ListTfidfVectorizer, self).fit(*args)
        logger.debug('There are {} vocabulary items in <{}>.'.format(len(self.vocabulary_), self))

        return ret
    #end def

    def fit_transform(self, X, *args, **kwargs):
        transformed = super(ListTfidfVectorizer, self).fit_transform(X)
        logger.debug('There {} vocabulary items in <{}>.'.format(len(self.vocabulary_), self))

        return transformed
    #end def
#end class


class ListHashingVectorizer(HashingVectorizer):
    def __init__(self, **kwargs):
        kwargs.setdefault('analyzer', ListNGramAnalyzer(**kwargs))
        super(ListHashingVectorizer, self).__init__(**kwargs)
    #end def
#end class


class CounterHashingVectorizer(HashingVectorizer):
    def __init__(self, **kwargs):
        kwargs.setdefault('analyzer', _counter_analyzer)
        super(CounterHashingVectorizer, self).__init__(**kwargs)
    #end def
#end class


class SpaceTokenizerTransformer(PureTransformer):
    def transform_one(self, text, **kwargs):
        return text.split()
#end class



from argparse import ArgumentParser
import logging
import sys

import numpy as np

from scipy.sparse import issparse

from tabulate import tabulate

from uriutils import URIFileType

from ..featurizers import load_featurizer, load_featurized, save_featurized
from ..utils import load_instances, load_dictionary_from_file, get_settings, parse_n_jobs, get_class_from_module_path

__all__ = []

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Featurize instances for ML classification.')
    parser.add_argument('featurizer_type', type=str, metavar='<featurizer_type>', nargs='?', default=None, help='Name of featurizer model to use.')
    parser.add_argument('-i', '--instances', type=URIFileType('r'), nargs='*', default=[], metavar='<instances>', help='List of instance files to featurize.')
    parser.add_argument('-o', '--output', type=URIFileType('wb'), metavar='<features_uri>', help='Save featurized instances here.')
    parser.add_argument('-s', '--settings', type=URIFileType(), metavar='<settings_uri>', help='Settings file to configure models.')
    parser.add_argument('--n-jobs', type=int, metavar='<N>', help='No. of processes to use during featurization.')
    parser.add_argument('--log-level', type=str, metavar='<log_level>', help='Set log level of logger.')
    parser.add_argument('--shuffle', action='store_true', help='Shuffle ordering of instances before writing them to file.')

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-f', '--fit', type=URIFileType('wb'), metavar='<featurizer>', help='Fit instances and save featurizer model file here.')
    group.add_argument('-t', '--featurize', type=URIFileType(), metavar='<featurizer_uri>', help='Use this featurizer to transform instances.')
    group.add_argument('-z', '--featurizer-info', type=URIFileType(), nargs='+', metavar='<featurizer_uri>', help='Display information about featurizer model.')
    group.add_argument('-x', '--features-info', type=URIFileType(), nargs='+', metavar='<featurized_uri>', help='Display information about featurized instance file.')
    group.add_argument('-v', '--verify', type=URIFileType(), metavar=('<featurizer_uri>', '<featurized_uri>'), nargs=2, help='Verify that the featurized instance file came from the same featurizer model.')
    A = parser.parse_args()

    file_settings = load_dictionary_from_file(A.settings) if A.settings else {}

    log_level = get_settings(key='log_level', sources=(A, 'env', file_settings), default='DEBUG').upper()
    log_format = get_settings(key='log_format', sources=(A, 'env', file_settings), default='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s')
    logging.basicConfig(format=log_format, level=logging.getLevelName(log_level))

    featurizer_type = get_settings(key='featurizer_type', sources=(A, 'env', file_settings))
    featurizer_parameters = get_settings(key='featurizer_parameters', sources=(file_settings, ), default={})
    featurizer_parameters['n_jobs'] = parse_n_jobs(get_settings(key='n_jobs', sources=(A, 'env', featurizer_parameters, file_settings), default=1))

    labels_field = get_settings(key='labels_field', sources=('env', file_settings), default='labels')
    logger.debug('Using "{}" for labels field.'.format(labels_field))

    if A.instances:
        X, Y_labels = list(map(np.array, zip(*load_instances(A.instances, labels_field=labels_field))))

    if A.fit:
        if not featurizer_type: parser.error('featurizer_type needs to be specified for fitting.'.format(featurizer_type))

        model_class = get_class_from_module_path(featurizer_type)
        if not model_class: parser.error('Unknown featurizer model "{}".'.format(featurizer_type))

        featurizer = model_class(**featurizer_parameters)
        X_featurized = featurizer.fit_transform(X, Y_labels)
        featurizer.save(A.fit)
        A.fit.close()

    elif A.featurize:
        featurizer = load_featurizer(A.featurize)
        if not A.output: logger.warning('No output file specified! The --featurize option is often used in conjunction with --output.')
        X_featurized = featurizer.transform(X)

    elif A.featurizer_info:
        if logger.getEffectiveLevel() > logging.INFO: logger.error('Log level should be at most INFO to view featurizer info.')
        for f in A.featurizer_info:
            featurizer = load_featurizer(f)
            logger.info('Featurizer info for <{}>:\n{}'.format(f.name, tabulate([
                ('Type', type(featurizer).__name__),
                ('UUID', featurizer.uuid),
                ('Repr', repr(featurizer)),
            ], headers=('Key', 'Value'), tablefmt='psql')))
        #end for

    elif A.features_info:
        if logger.getEffectiveLevel() > logging.INFO: logger.error('Log level should be at most INFO to view featurized instances info.')
        for f in A.features_info:
            featurizer_uuid, X_featurized, featurized_at = load_featurized(f, ('featurizer_uuid', 'X_featurized', 'featurized_at'))
            info_table = [
                ('UUID', featurizer_uuid),
                ('Featurized at', str(featurized_at) + ' UTC'),
            ]
            feature_count = X_featurized.shape[1] if X_featurized.ndim >= 2 else 1
            if issparse(X_featurized):
                info_table += [
                    ('Matrix type', 'Sparse'),
                    ('Matrix dimensions', '{}x{} ({:,} active features)'.format(X_featurized.shape[0], feature_count, len(X_featurized.data))),
                    ('Matrix sparsity', '{:.3f}%'.format(len(X_featurized.data) / (X_featurized.shape[0] * feature_count) * 100.0)),
                ]
            else:
                info_table += [('Matrix type', 'Dense'), ('Matrix dimensions', '{}x{}'.format(X_featurized.shape[0], feature_count))]
            #end if

            logger.info('Featurizer info for <{}>:\n{}'.format(f.name, tabulate(info_table, headers=('Key', 'Value'), tablefmt='psql')))
        #end for

    elif A.verify:
        featurizer = load_featurizer(A.verify[0])
        featurizer_uuid = load_featurized(A.verify[1], ('featurizer_uuid',))[0]
        if featurizer_uuid == featurizer.uuid: logger.info('UUID match OK.')
        else:
            logger.error('UUID mismatch! Featurizer UUID {} != {}'.format(featurizer.uuid, featurizer_uuid))
            sys.exit(-1)
    #end if

    if A.fit or A.featurize:
        logger.debug('Feature matrix has dimensions {} ({:,} active features).'.format('x'.join(map(str, X_featurized.shape)), len(X_featurized.data)))

        if A.output:
            if A.shuffle:
                shuffled_indexes = np.random.permutation(X.shape[0])
                logger.info('Featurized instances shuffled.')
            else:
                shuffled_indexes = np.arange(X.shape[0])
            #end if

            id_key = None
            for key in ['_id', 'id', 'id_', 'uuid']:
                if key in X[0]:
                    id_key = key
                    break
                #end if
            #end for
            if not id_key: raise TypeError('Unable to find ID key in instances.')
            else: logger.info('Using "{}" as key for ID field.'.format(id_key))

            X_meta = np.array([dict(id=X[i][id_key]) for i in shuffled_indexes], dtype=np.object)
            save_featurized(A.output, X_featurized=X_featurized[shuffled_indexes, ...], Y_labels=Y_labels[shuffled_indexes], X_meta=X_meta, featurizer_uuid=featurizer.uuid_)
            A.output.close()
        #end if
    #end if
#end def


if __name__ == '__main__': main()



__all__ = ['SpacyNLPProcessor', 'SpacyDocSerializer', 'NounPhraseExtractor', 'NamedEntityExtractor', 'TextRankerKeywordExtractor']

from itertools import combinations, product
import json
import logging
import os
import re

try: import networkx as nx
except ImportError: nx = None

try:
    import spacy
    from spacy import symbols
except ImportError:
    spacy = None
    symbols = None
#end try

from .base import PureTransformer

logger = logging.getLogger(__name__)


class SpacyNLPProcessor(PureTransformer):
    def __init__(self, model_name='en', spacy_args={}, use_tagger=True, use_parser=True, use_entity=True, batch_size=128, n_jobs=None, generator=False, **kwargs):
        kwargs.setdefault('nparray', False)
        super(SpacyNLPProcessor, self).__init__(**kwargs)

        if spacy is None:
            raise ImportError('You need to install the spacy NLP package.')

        if n_jobs is None: self.n_jobs = int(os.environ.get('N_JOBS', 1))
        else: self.n_jobs = n_jobs

        self.batch_size = batch_size
        self.generator = generator

        self.spacy_args = spacy_args
        self.model_name = model_name
        self.use_tagger = use_tagger
        self.use_parser = use_parser
        self.use_entity = use_entity

        self.nlp = self._setup_nlp(model_name, spacy_args, use_tagger, use_parser, use_entity)
    #end def

    def _setup_nlp(self, model_name='en', spacy_args={}, use_tagger=True, use_parser=True, use_entity=True):
        nlp = spacy.load(model_name, **spacy_args)
        if 'pipeline' not in spacy_args and 'create_pipeline' not in spacy_args:
            pipeline = []
            if use_tagger: pipeline.append(nlp.tagger)
            if use_parser: pipeline.append(nlp.parser)
            if use_entity: pipeline.append(nlp.entity)

            nlp.pipeline = pipeline
        #end if

        logger.debug('Loaded spacy model {}'.format(json.dumps(nlp.meta)))

        return nlp
    #end def

    def _transform(self, X_texts, **kwargs):
        if self.generator:
            return (doc for doc in self.nlp.pipe(X_texts, batch_size=self.batch_size, n_threads=self.n_jobs))
        return [doc for doc in self.nlp.pipe(X_texts, batch_size=self.batch_size, n_threads=self.n_jobs)]
    #end def

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop('nlp', None)

        return state
    #end def

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.nlp = self._setup_nlp(self.model_name, self.spacy_args, self.use_tagger, self.use_parser, self.use_entity)
    #end def
#end def


class SpacyDocSerializer(PureTransformer):
    def __init__(self, **kwargs):
        kwargs.setdefault('nparray', False)
        super(SpacyDocSerializer, self).__init__(**kwargs)
    #end def

    def transform_one(self, doc, **kwargs):
        tokens = {}
        for tok in doc:
            tokens[tok.i] = dict(idx=tok.idx, ent_type=tok.ent_type_, ent_iob=tok.ent_iob_, lemma=tok.lemma_, text=tok.text, pos=tok.pos_, tag=tok.tag_, head=tok.head.i, dep=tok.dep_)
        #end for

        sentences = [dict(tokens=[tok.i for tok in sent]) for sent in doc.sents]
        entities = [dict(tokens=[tok.i for tok in ent], label=ent.label_) for ent in doc.ents]
        noun_chunks = [dict(tokens=[tok.i for tok in chunk]) for chunk in doc.noun_chunks]

        d = dict(length=len(doc), text=doc.text, tokens=tokens, sentences=sentences, entities=entities, noun_chunks=noun_chunks)

        return d
    #end def
#end class


class NamedEntityExtractor(PureTransformer):
    def __init__(self, entity_labels=None, **kwargs):
        kwargs.setdefault('nparray', False)
        super(NamedEntityExtractor, self).__init__(**kwargs)

        self.entity_labels = entity_labels
    #end def

    def transform_one(self, doc, **kwargs):
        entities = [ent for ent in doc.ents if self.entity_labels is None or ent.label in self.entity_labels or ent.label_ in self.entity_labels]

        return entities
    #end def
#end class


class NounPhraseExtractor(PureTransformer):
    # def __init__(self, np_rules=[r'(N|J)'], **kwargs):
    def __init__(self, np_rules=[r'((J|X|N|0)(J|C|X|0|N)*(N))'], **kwargs):
        kwargs.setdefault('nparray', False)
        super(NounPhraseExtractor, self).__init__(**kwargs)

        self.POS_TO_CHAR = {
            symbols.ADJ: 'J',
            symbols.ADP: 'P',
            symbols.ADV: 'R',
            symbols.AUX: 'A',
            symbols.CONJ: 'C',
            symbols.CCONJ: 'C',
            symbols.DET: 'D',
            symbols.INTJ: 'I',
            symbols.NOUN: 'N',
            symbols.NUM: '0',
            symbols.PART: 'T',
            symbols.PRON: 'O',
            symbols.PROPN: 'N',
            symbols.PUNCT: '.',
            symbols.SCONJ: 'C',
            symbols.SYM: '#',
            symbols.VERB: 'V',
            symbols.X: 'X',
            symbols.SPACE: '.',
            symbols.EOL: '.',
        }

        self.np_rules = []
        for rule in np_rules:
            if isinstance(rule, str): self.np_rules.append(re.compile(rule))
            else: self.np_rules.append(rule)
        #end for
    #end def

    def transform_one(self, doc, **kwargs):
        chunks = []
        for sent in doc.sents:
            pos_tags = ''.join(self.POS_TO_CHAR[tok.pos] for tok in sent)

            for rule in self.np_rules:
                for m in rule.finditer(pos_tags):

                    chunks.append(sent[m.span(1)[0]:m.span(1)[1]])
                    # print(sent[m.span(1)[0]:m.span(1)[1]], pos_tags[m.span(1)[0]:m.span(1)[1]])
                #end for
            #end for
        #end for

        return chunks
    #end def
#end class


class TextRankerKeywordExtractor(PureTransformer):
    def __init__(self, use_lemma=True, coocurrence_window_size=8, **kwargs):
        kwargs.setdefault('nparray', False)
        super(TextRankerKeywordExtractor, self).__init__(**kwargs)

        if nx is None: raise ImportError('You need to install networkx for the TextRankerKeywordExtractor.')

        self.use_lemma = use_lemma
        self.coocurrence_window_size = coocurrence_window_size
    #end def

    def transform_one(self, phrases, **kwargs):
        vocabulary = []
        vocab_map = {}

        for phrase in phrases:
            normalized = self.normalize_phrase(phrase)
            if normalized in vocab_map:
                vocab_map[normalized].append(phrase)
            else:
                vocab_map[normalized] = [phrase]
                vocabulary.append(normalized)
            #end if
        #end for

        vocabulary.sort()
        # vocab_map_index = dict((normalized, i) for i, normalized in enumerate(vocabulary))
        N = len(vocabulary)

        # Build the graph
        gr = nx.Graph()
        gr.add_nodes_from(range(N))
        for i, j in combinations(range(N), 2):
            phrases_i = vocab_map[vocabulary[i]]
            phrases_j = vocab_map[vocabulary[j]]
            distance = self.compute_cooccurrence_distance(phrases_i, phrases_j)
            if distance <= self.coocurrence_window_size:
                gr.add_edge(i, j)
        #end for

        calculated_page_rank = nx.pagerank(gr)
        keyphrase_indexes = sorted(calculated_page_rank.keys(), key=calculated_page_rank.get, reverse=True)
        keyphrases = []
        for i in keyphrase_indexes:
            keyphrases.append((vocab_map[vocabulary[i]][0], calculated_page_rank[i]))
        #end for

        return keyphrases
    #end def

    def compute_cooccurrence_distance(self, phrases_i, phrases_j):
        # for p_i, p_j in product(phrases_i, phrases_j):
        #     print(p_i, 'x', p_j, abs(p_i[0].i - p_j[0].i))

        return min(abs(p_i[0].i - p_j[0].i) for p_i, p_j in product(phrases_i, phrases_j))
    #end def

    def normalize_phrase(self, phrase):
        if self.use_lemma:
            return ' '.join(tok.lemma_ for tok in phrase)
        return ' '.join(tok.lower_ for tok in phrase)
    #end def
#end class


def main():
    logging.basicConfig(format='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s', level=logging.DEBUG)

    nlp_processor = SpacyNLPProcessor()
    np_extractor = NounPhraseExtractor()
    textrank_extractor = TextRankerKeywordExtractor()

    text = """
    Bertram "Bert" Thomas Combs (August 13, 1911 – December 4, 1991) was a jurist and politician from the U.S. state of Kentucky. After serving on the Kentucky Court of Appeals, he was elected the 50th Governor of Kentucky in 1959 on his second run for the office. Following his gubernatorial term, he was appointed to the Sixth Circuit Court of Appeals by President Lyndon B. Johnson, serving from 1967 to 1970.

    Combs rose from poverty in his native Clay County to obtain a law degree from the University of Kentucky and open a law practice in Prestonsburg. He was decorated for prosecuting Japanese war criminals before military tribunals following World War II, then returned to Kentucky and his law practice. In 1951, Governor Lawrence Wetherby appointed him to fill a vacancy on the Kentucky Court of Appeals. Later that year, he was elected to a full term on the court, defeating former governor and judge Simeon S. Willis. Kentucky's Democratic Party had split into two factions by 1955 when Earle C. Clements, the leader of one faction, chose Combs to challenge former governor and U.S. Senator A. B. "Happy" Chandler, who headed the other, in the upcoming gubernatorial primary. Combs' uninspiring speeches and candidness about the need for more state revenue cost him the primary election. Chandler, who went on to reclaim the governorship, had promised that he would not need to raise taxes to meet the state's financial obligations, but ultimately he did so. This damaged Chandler's credibility and left Combs looking courageous and honest in the eyes of the electorate. Consequently, in 1959 Combs was elected governor, defeating Lieutenant Governor Harry Lee Waterfield, Chandler's choice to succeed him in office, in the primary. Early in his term, Combs secured passage of a three-percent sales tax to pay a bonus to the state's military veterans. Knowing a tax of one percent would have been sufficient, he used the excess revenue to enact a system of reforms, including expansion of the state's highway and state park systems. He also devoted much of the surplus to education.
    """

    for doc_chunks in textrank_extractor.transform(np_extractor.transform(nlp_processor.transform([text]))):
        print(doc_chunks)
#end def


if __name__ == '__main__': main()



__all__ = ['ObjectFeatureUnion', 'ListConcatTransformer']

from collections import OrderedDict

import numpy as np

from . import PureTransformer


class ObjectFeatureUnion(PureTransformer):
    """Extract a given key from dictionary object."""

    def __init__(self, *args, as_dict=True, **kwargs):
        kwargs.setdefault('nparray', True)
        kwargs.setdefault('nparray_dtype', np.object)
        super(ObjectFeatureUnion, self).__init__(**kwargs)

        self.as_dict = as_dict
        self.steps = OrderedDict()

        for (k, step) in args:
            self.steps[k] = step
    #end def

    def fit(self, *args, **fit_params):
        for k, step in self.steps.items():
            step.fit(*args, **fit_params)

        return self
    #end def

    def _transform(self, X, y=None, **kwargs):
        if self.as_dict:
            transformed = []
            first = True
            # zip(*[step.transform(X, **kwargs) for k, step in self.steps.items()])
            for k, step in self.steps.items():
                step_transformed = step.transform(X, **kwargs)
                for i, t in enumerate(step_transformed):
                    if first:
                        transformed.append({k: t})
                    else:
                        transformed[i][k] = t
                #end for

                first = False
            #end for

        else:
            transformed = []
            first = True
            for k, step in self.steps.items():
                step_transformed = step.transform(X, y=y, **kwargs)
                for i, t in enumerate(step_transformed):
                    if first:
                        transformed.append([t])
                    else:
                        transformed[i].append(t)
                #end for
            #end for
        #end if

        return transformed
    #end def
#end class


class ListConcatTransformer(PureTransformer):
    """Concatenate two lists for each instance."""

    def __init__(self, steps=[], **kwargs):
        kwargs.setdefault('nparray', False)
        super(ListConcatTransformer, self).__init__(**kwargs)

        self.steps = steps
    #end def

    def fit(self, *args, **kwargs):
        for step in self.steps:
            step.fit(*args, **kwargs)

        return self
    #end def

    def _transform(self, X, *args, **kwargs):
        N = len(X)
        transformed = [[] for i in range(N)]
        for step in self.steps:
            transformed_step = step.transform(X, *args, **kwargs)
            assert len(transformed_step) == N
            for j in range(N):
                transformed[j] += transformed_step[j]
        #end for

        return transformed
    #end def
#end class



__all__ = ['parse_n_jobs']

import logging
from multiprocessing import cpu_count
import re

logger = logging.getLogger(__name__)


def parse_n_jobs(s):
    n_jobs = None
    N = cpu_count()

    if isinstance(s, int): n_jobs = s

    elif isinstance(s, float): n_jobs = int(s)

    elif isinstance(s, str):
        m = re.match(r'(\d*(?:\.\d*)?)?(\s*\*?\s*n)?$', s.strip())
        if m is None: raise ValueError('Unable to parse n_jobs="{}"'.format(s))

        k = float(m.group(1)) if m.group(1) else 1
        if m.group(2): n_jobs = k * N
        elif k < 1: n_jobs = k * N
        else: n_jobs = int(k)

    else: raise TypeError('n_jobs argument must be of type str, int, or float.')

    n_jobs = int(n_jobs)
    if n_jobs <= 0:
        logger.warning('n_jobs={} is invalid. Setting n_jobs=1.'.format(n_jobs))
        n_jobs = 1
    #end if

    if n_jobs > N:
        logger.warning('n_jobs={} is > number of available CPUs ({}).'.format(s, N))

    return int(n_jobs)
#end def



__all__ = ['HyperparameterGridsearchMixin']

import logging

from joblib import Parallel, delayed

from sklearn.metrics import accuracy_score
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection import train_test_split

from ycml.utils import get_class_from_module_path

logger = logging.getLogger(__name__)


class HyperparameterGridsearchMixin(object):
    def __init__(self, classifier='sklearn.svm.SVC', classifier_args={}, param_grid={}, metric=accuracy_score, validation_size=0.2, **kwargs):
        self.classifier = classifier
        self.classifier_args = classifier_args
        self.param_grid = param_grid
        self.validation_size = validation_size

        self.metric = metric
        if isinstance(metric, str):
            if metric.startswith('lambda'): self.metric = eval(metric)
            else: self.metric = get_class_from_module_path(metric)
        #end if
    #end def

    def fit_binarized(self, X_featurized, Y_binarized, validation_data=None, **kwargs):
        klass = get_class_from_module_path(self.classifier)

        if validation_data is None:  # use 0.2 for validation data
            X_train, X_validation, Y_train, Y_validation = train_test_split(X_featurized, Y_binarized, test_size=self.validation_size)
            logger.info('Using {} of training data ({} instances) for validation.'.format(self.validation_size, Y_validation.shape[0]))
        else:
            X_train, X_validation, Y_train, Y_validation = X_featurized, validation_data[0], Y_binarized, validation_data[1]
        #end if

        best_score, best_param = 0.0, None

        if self.n_jobs > 1: logger.info('Performing hyperparameter gridsearch in parallel using {} jobs.'.format(self.n_jobs))
        else: logger.debug('Performing hyperparameter gridsearch in parallel using {} jobs.'.format(self.n_jobs))

        param_scores = Parallel(n_jobs=self.n_jobs)(delayed(_fit_classifier)(klass, self.classifier_args, param, self.metric, X_train, Y_train, X_validation, Y_validation) for param in ParameterGrid(self.param_grid))

        best_param, best_score = max(param_scores, key=lambda x: x[1])
        logger.info('Best scoring param is {} with score {}.'.format(best_param, best_score))

        classifier_args = {}
        classifier_args.update(self.classifier_args)
        classifier_args.update(best_param)
        self.classifier_ = klass(**classifier_args)
        logger.info('Fitting final model <{}> on full data with param {}.'.format(self.classifier_, best_param))
        self.classifier_.fit(X_featurized, Y_binarized)

        return self
    #end def
#end class


def _fit_classifier(klass, classifier_args, param, metric, X_train, Y_train, X_validation, Y_validation):
    local_classifier_args = {}
    local_classifier_args.update(classifier_args)
    local_classifier_args.update(param)

    classifier = klass(**local_classifier_args).fit(X_train, Y_train)

    Y_predict = classifier.predict(X_validation)
    score = metric(Y_validation, Y_predict)
    logger.info('<{}> with param {} has micro F1 of {}.'.format(classifier, param, score))

    return (param, score)
#end def



__all__ = ['BinaryLabelsClassifier', 'MultiLabelsClassifier', 'MulticlassLabelsClassifier', 'filter_labels']

import logging

import numpy as np

from sklearn.preprocessing import MultiLabelBinarizer

from .base import BaseClassifier

logger = logging.getLogger(__name__)


class LabelsClassifierMixin(BaseClassifier):
    def __init__(self, include=[], exclude=[], **kwargs):
        super(LabelsClassifierMixin, self).__init__(**kwargs)

        self.exclude = set(exclude)
        self.include = set(include)
    #end def

    def _fit(self, X, Y_labels, *, validation_data=None, binarize_args={}, fit_args={}, **kwargs):
        Y_binarized = self.binarize_labels(Y_labels, **binarize_args, **kwargs)
        if validation_data is not None:
            X_validation, Y_validation = validation_data
            validation_data = (X_validation, self.binarize_labels(Y_validation, **binarize_args, **kwargs))
            logger.debug('Binarized validation labels.')
        #end if

        return self.fit_binarized(X, Y_binarized, validation_data=validation_data, **fit_args, **kwargs)
    #end def

    def fit_binarized(self, X_featurized, Y_binarized, **kwargs): raise NotImplementedError('fit_binarized is not implemented.')

    def binarize_labels(self, Y_labels, **kwargs): raise NotImplementedError('binarize_labels is not implemented.')

    def binarize_dicts(self, Y_dicts, *, default=0.0, **kwargs): raise NotImplementedError('binarize_dicts is not implemented.')

    def unbinarize_labels(self, Y_proba, *, epsilon=1e-5, to_dict=False, astype=float, **kwargs):
        assert len(self.classes_) == Y_proba.shape[1]
        unbinarized = np.empty(Y_proba.shape[0], dtype=np.object)
        for i in range(Y_proba.shape[0]):
            if to_dict: unbinarized[i] = dict((self.classes_[j], astype(Y_proba[i, j])) for j in range(Y_proba.shape[1]) if Y_proba[i, j] > epsilon)
            else: unbinarized[i] = [self.classes_[j] for j in range(Y_proba.shape[1]) if Y_proba[i, j] > epsilon]
        #end for

        return unbinarized
    #end def
#end def


class BinaryLabelsClassifier(LabelsClassifierMixin):
    def __init__(self, pos_label, *, not_pos_label=None, **kwargs):
        super(BinaryLabelsClassifier, self).__init__(**kwargs)

        self.pos_label = pos_label
        self.not_pos_label = 'not ' + pos_label if not_pos_label is None else not_pos_label
    #end def

    def binarize_labels(self, Y_labels, **kwargs):
        Y_binarized = np.zeros((Y_labels.shape[0], 2))
        for i in range(Y_labels.shape[0]):
            if self.pos_label in Y_labels[i]: Y_binarized[i, 1] = 1
            else: Y_binarized[i, 0] = 1
        #end for

        return Y_binarized
    #end def

    def binarize_dicts(self, Y_dicts, *, default=0.0, **kwargs):
        Y_binarized = np.zeros((Y_dicts.shape[0], 2))
        for i in range(Y_dicts.shape[0]):
            Y_binarized[i, 1] = Y_dicts[i].get(self.pos_label, default)
        Y_binarized[:, 0] = 1.0 - Y_binarized[:, 1]

        return Y_binarized
    #end def

    @property
    def classes_(self):
        return [self.not_pos_label, self.pos_label]
#end def


class MultiLabelsClassifier(LabelsClassifierMixin):
    def _fit(self, X, Y_labels, **kwargs):
        Y_labels_filtered = filter_labels(Y_labels, include=self.include, exclude=self.exclude)
        self.label_binarizer_ = MultiLabelBinarizer(sparse_output=False).fit(Y_labels_filtered)
        logger.info('{} labels found in training instances.'.format(len(self.classes_)))

        if not len(self.classes_): raise ValueError('There are no labels available for fitting model.')

        return super(MultiLabelsClassifier, self)._fit(X, Y_labels_filtered, **kwargs)
    #end def

    def binarize_labels(self, Y_labels, **kwargs):
        classes_ = set(self.classes_)

        return self.label_binarizer_.transform((filter(classes_.__contains__, labels) for labels in Y_labels))
    #end def

    def binarize_dicts(self, Y_dicts, *, default=0.0, **kwargs):
        binarized = np.fill((Y_dicts.shape[0], len(self.classes_)), default, dtype=np.float)
        classes_map = dict((c, i) for i, c in enumerate(self.classes_))

        for i in range(Y_dicts.shape[0]):
            d = Y_dicts[i]
            for k, p in d.items():
                try: binarized[i, classes_map[k]] = p
                except IndexError: pass
            #end for
        #end for

        return binarized
    #end def

    @property
    def classes_(self):
        return self.label_binarizer_.classes_
    #end def
#end def


class MulticlassLabelsClassifier(MultiLabelsClassifier):
    def _fit(self, X, Y_labels, **kwargs):
        Y_labels_filtered = self._filter_and_check_labels(Y_labels)
        if any(len(Y_labels_filtered[i]) == 0 for i in range(Y_labels_filtered.shape[0])):
            logger.warning('Some of Y_labels have no labels after filtering but this is a multiclass classifier. A `<none>` label will be created.')

            for i in range(Y_labels_filtered.shape[0]):
                if not Y_labels_filtered[i]:
                    Y_labels_filtered[i] = ['<none>']
        #end if

        return super(MulticlassLabelsClassifier, self)._fit(X, Y_labels_filtered, **kwargs)
    #end def

    def predict(self, X_featurized, **kwargs):
        Y_proba = self.predict_proba(X_featurized, **kwargs)
        Y_predict_binarized = np.zeros(Y_proba.shape)
        for i in range(X_featurized.shape[0]):
            j = np.argmax(Y_proba[i, :])
            Y_predict_binarized[i, j] = 1
        #end for

        return Y_predict_binarized
    #end def

    def predict_and_proba(self, X_featurized, **kwargs):
        Y_proba = super(MulticlassLabelsClassifier, self).predict_proba(X_featurized, **kwargs)

        Y_predict_binarized = np.zeros(Y_proba.shape)
        for i in range(X_featurized.shape[0]):
            j = np.argmax(Y_proba[i, :])
            Y_predict_binarized[i, j] = 1
        #end for

        return Y_proba, Y_predict_binarized
    #end def

    def binarize_labels(self, Y_labels, **kwargs):
        Y_labels_filtered = self._filter_and_check_labels(Y_labels)

        return super(MulticlassLabelsClassifier, self).binarize_labels(Y_labels_filtered, **kwargs)
    #end def

    def multiclassify_labels(self, Y_labels, **kwargs):
        Y_binarized = self.binarize_labels(Y_labels, **kwargs)
        return MulticlassLabelsClassifier.multilabel_to_multiclass(Y_binarized)
    #end def

    @classmethod
    def multilabel_to_multiclass(cls, Y_multilabel):
        Y_multiclass = np.zeros(Y_multilabel.shape[0])
        for j in range(Y_multilabel.shape[1]):
            Y_multiclass[Y_multilabel[:, j] > 0] = j

        return Y_multiclass
    #end def

    @classmethod
    def multiclass_to_multilabel(cls, Y_multiclass, n_classes=None):
        if n_classes is None: n_classes = np.max(Y_multiclass)
        Y_multilabel = np.zeros((Y_multiclass.shape[0], n_classes))
        for j in range(Y_multilabel.shape[1]):
            Y_multilabel[Y_multiclass == j, j] = 1

        return Y_multiclass
    #end def

    def _filter_and_check_labels(self, Y_labels):
        Y_labels_filtered = filter_labels(Y_labels, include=self.include, exclude=self.exclude)

        if any(len(Y_labels[i]) > 1 for i in range(Y_labels.shape[0])):
            logger.warning('Some of Y_labels contain more than 1 labels but this is a multiclass classifier. Only the first labels will be used.')

        return np.array([[Y_labels_filtered[i][0]] if Y_labels_filtered[i] else [] for i in range(Y_labels_filtered.shape[0])])
    #end def

    @property
    def classes_(self):
        return self.label_binarizer_.classes_
    #end def
#end def


def filter_labels(Y_labels, *, include=[], exclude=[]):
    if not exclude and not include: return Y_labels
    if include: logger.debug('Included labels: {}'.format(', '.join(sorted(include))))
    if exclude: logger.debug('Excluded labels: {}'.format(', '.join(sorted(exclude))))

    Y_labels_filtered = []
    removed_labels = 0
    for i in range(Y_labels.shape[0]):
        Y_labels_filtered.append([l for l in Y_labels[i] if (l in include or not include) and l not in exclude])
        removed_labels += len(Y_labels[i]) - len(Y_labels_filtered[i])
    #end for
    Y_labels_filtered = np.array(Y_labels_filtered, dtype=np.object)
    if removed_labels: logger.info('{} label-instances removed from the data.'.format(removed_labels))

    return Y_labels_filtered
#end def



__all__ = ['PureTransformer', 'identity']

import logging

import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin

from ..utils import Timer

logger = logging.getLogger(__name__)


# Helper class. A transformer that only does transformation and does not need to fit any internal parameters.
class PureTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, nparray=True, nparray_dtype=None, generator=False, **kwargs):
        super(PureTransformer, self).__init__()

        self.nparray = nparray
        self.nparray_dtype = nparray_dtype
        self.generator = generator

        if self.nparray and self.generator:
            raise ValueError('nparray and generator option cannot both be True.')
    #end def

    def fit(self, *args, **fit_params):
        return self

    def transform(self, X, y=None, **kwargs):
        timer = Timer()
        transformed = self._transform(X, y=y, **kwargs)
        if self.nparray:
            nparray_dtype = getattr(self, 'nparray_dtype', None)
            if nparray_dtype:
                transformed = np.array(transformed, dtype=nparray_dtype)
            else:
                transformed = np.array(transformed)
                if transformed.ndim == 1:
                    transformed = transformed.reshape(transformed.shape[0], 1)
            #end if
        #end if

        logger.debug('Done <{}> transformation {}.'.format(type(self).__name__, timer))

        return transformed
    #end def

    def _transform(self, X, y=None, **kwargs):
        if self.generator:
            return (self.transform_one(row, **kwargs) for row in X)

        return [self.transform_one(row, **kwargs) for row in X]
    #end def

    def transform_one(self, x, **kwargs):
        raise NotImplementedError('transform_one method needs to be implemented.')
#end class


def identity(x): return x



from argparse import ArgumentParser
import logging
import random

from uriutils import URIFileType

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Script to partition file by lines.')
    parser.add_argument('-i', '--instances', type=URIFileType('r'), nargs='+', metavar='<instances>', help='List of instance files to partition.')
    parser.add_argument('-s', '--sizes', type=float, nargs='+', required=True, default=[], metavar='<N>', help='Proportions/Number of instances in each partition.')
    parser.add_argument('--shuffle', action='store_true', default=False, help='Whether to shuffle the instances before partitioning.')
    parser.add_argument('-p', '--prefix', type=str, default='partition', metavar='<prefix>', help='Prefix filename to use when saving partitions.')
    parser.add_argument('-o', '--output', type=URIFileType('w'), nargs='*', default=[], metavar='<output>', help='Save partitioned instances here.')
    A = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s', level=logging.INFO)

    if A.prefix != 'partition' and A.output: parser.error('--prefix and --output cannot be used together.')
    if A.output: output_files = A.output
    else: output_files = [open(A.prefix + '{:02d}'.format(i), 'w') for i, s in enumerate(A.sizes)]
    if len(A.sizes) != len(output_files): parser.error('Number of arguments in --sizes and --output mismatch.')

    instances = []
    for f in A.instances:
        lines = [line.strip() for line in f]
        instances += lines
        logger.info('{} lines read from instance file <{}>.'.format(len(lines), f.name))
    #end for
    N = len(instances)
    logger.info('{} instances found.'.format(N))

    if A.shuffle:
        random.shuffle(instances)
        logger.info('Instances shuffled.')
    #end if

    Z = sum(A.sizes)
    N_partitions = len(A.sizes)

    output_lines = [int(s / Z * N) for s in A.sizes]
    leftovers = N - sum(output_lines)
    for i in range(leftovers):
        output_lines[i % N_partitions] += 1
    assert sum(output_lines) == N

    cur = 0
    for num_instances, g in zip(output_lines, output_files):
        for i in range(num_instances):
            g.write(instances[cur])
            g.write('\n')
            cur += 1
        #end for
        g.close()

        logger.info('Wrote {} instances to <{}>.'.format(num_instances, g.name))
    #end for
#end def


if __name__ == '__main__': main()



from argparse import ArgumentParser
import logging

from uriutils import URIFileType

from ..featurizers import load_featurized
from ..utils import load_dictionary_from_file, get_settings, get_class_from_module_path

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Classify instances using ML classifier.')
    parser.add_argument('--log-level', type=str, metavar='<log_level>', help='Set log level of logger.')
    parser.add_argument('-s', '--settings', type=URIFileType(), metavar='<settings_file>', help='Settings file to configure models.')
    parser.add_argument('--n-jobs', type=int, metavar='<N>', help='No. of processor cores to use.')

    parser.add_argument('classifier_type', type=str, metavar='<classifier_type>', nargs='?', help='Type of classifier model to fit.')
    parser.add_argument('-f', '--featurized', type=URIFileType(), metavar='<featurized>', help='Fit model on featurized instances.')
    parser.add_argument('-o', '--output', type=URIFileType('wb'), metavar='<classifier_file>', help='Save trained classifier model here.')

    parser.add_argument('-r', '--resume', type=str, metavar='<param>', nargs='+', help='Resume training from given file (takes multiple arguments depending on classifier).')
    parser.add_argument('-v', '--validation-data', type=URIFileType(), metavar='<featurized>', help='Use this as validation set instead of system defined one.')

    A = parser.parse_args()

    file_settings = load_dictionary_from_file(A.settings) if A.settings else {}

    log_level = get_settings(key='log_level', sources=(A, 'env', file_settings), default='DEBUG').upper()
    log_format = get_settings(key='log_format', sources=(A, 'env', file_settings), default='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s')
    logging.basicConfig(format=log_format, level=logging.getLevelName(log_level))

    classifier_type = get_settings(key='classifier_type', sources=(A, 'env', file_settings))
    classifier_parameters = get_settings((file_settings, 'classifier_parameters'), default={})
    classifier_parameters['n_jobs'] = get_settings(key='n_jobs', sources=(A, 'env', classifier_parameters, file_settings), default=1)

    classifier_class = get_class_from_module_path(classifier_type)
    if not classifier_class: parser.error('Unknown classifier name "{}".'.format(classifier_type))

    X_featurized, Y_labels = load_featurized(A.featurized, keys=('X_featurized', 'Y_labels'))

    kwargs = dict(fit_args={})
    if A.validation_data:
        X_validation, Y_validation = load_featurized(A.validation_data, keys=('X_featurized', 'Y_labels'))
        kwargs['validation_data'] = (X_validation, Y_validation)
    #end if

    if A.resume: kwargs['fit_args']['resume'] = A.resume

    classifier = classifier_class(**classifier_parameters).fit(X_featurized, Y_labels, **kwargs)

    if A.output: classifier.save(A.output)
#end def


if __name__ == '__main__': main()



_all__ = ['Timer']

import time


class Timer(object):
    def __init__(self, display_format='(took {elapsed:.3f} {unit})', display_unit='seconds'):
        self.start_time = time.time()
        self.display_unit = display_unit
        self.display_format = display_format
    #end def

    @property
    def elapsed_time(self):
        return time.time() - self.start_time

    def reset(self):
        self.start_time = time.time()

    def __str__(self):
        if self.display_unit == 'seconds': elapsed = self.elapsed_time
        elif self.display_unit == 'milliseconds': elapsed = self.elapsed_time * 1000.0

        return self.display_format.format(elapsed=elapsed, unit=self.display_unit)
    #end def
#end class



from __future__ import print_function, unicode_literals

from argparse import ArgumentParser
import logging
import os
import re
import time

from uriutils import URIFileType

logger = logging.getLogger(__name__)


VM_REGEX = re.compile(r'(Vm\w+)\:\s+(\d+) kB')
NAME_REGEX = re.compile(r'(Name)\:\s+(.+)$')
TRACKED_STATUS = ['VmPeak', 'VmRSS', 'VmSize']


def main():
    parser = ArgumentParser(description='Script to track memory usage of processes.')
    parser.add_argument('pid', type=int, metavar='<pid>', nargs='+', help='Process IDs to track.')
    parser.add_argument('--rate', type=float, metavar='<rate>', default=30, help='Sample every rate seconds.')
    parser.add_argument('--save', type=URIFileType('w'), metavar='<file>', help='Save sampled information to file.')
    A = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s [%(name)s] %(levelname)s: %(message)s', level=logging.DEBUG)

    logger.info('Collecting memory information for PIDs {} at a rate of {} seconds.'.format(A.pid, A.rate))
    ignored_pid = set()
    if A.save: print('#', '\t'.join(['Timestamp', 'Pid-Name', 'Attribute', 'kB'], file=A.save))
    last_info = time.time()
    sample_count = 0
    have_data = True
    while True:
        timestamp = '{:.0f}'.format(time.time())
        for pid in A.pid:
            if pid in ignored_pid: continue
            status_file = os.path.join('/proc', str(pid), 'status')
            if not os.path.isfile(status_file):
                logger.warning('Process {} has disappeared. Will not track it anymore.'.format(pid))
                ignored_pid.add(pid)
                continue
            #end if

            status = {}
            with open(os.path.join('/proc', str(pid), 'status')) as f:
                for line in f:
                    m = NAME_REGEX.match(line.strip())
                    if m: status[m.group(1)] = m.group(2)
                    m = VM_REGEX.match(line)
                    if m: status[m.group(1)] = m.group(2)
                #end for
            #end with
            sample_count += 1

            proc_key = '{}-{}'.format(pid, status['Name'])
            s = ['Process={}'.format(proc_key)]
            for k in TRACKED_STATUS:
                if A.save:
                    print('\t'.join([timestamp, proc_key, k, status[k]]), file=A.save)
                s.append('{}={}'.format(k, convert_memory(status[k])))
            #end for
            logger.info('; '.join(s))
            have_data = True
        #end for

        if have_data and time.time() - last_info > A.rate * 10:
            logger.info('Collected {} samples so far...'.format(sample_count))
            last_info = time.time()
            have_data = False
        #end if

        time.sleep(A.rate)
    #end while
#end def


def convert_memory(kb_size):
    converted = float(kb_size)
    converted_unit = 'kB'

    for unit in ['MB', 'GB', 'TB', 'PB']:
        if converted > 1024:
            converted /= 1024.0
            converted_unit = unit
        else: break
    #end for

    return '{:.3f}{}'.format(converted, converted_unit)
#end def


if __name__ == '__main__': main()



"""
python -m helloworld.generate_data data/train.json.gz
python -m helloworld.generate_data data/evaluate.json.gz -N 200
"""

from argparse import ArgumentParser
from itertools import combinations
import json
import logging
from random import choice
from uuid import uuid4

from ycml.utils import URIFileType

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Generates a sample data set for demo purposes.')
    parser.add_argument('output', type=URIFileType('w'), metavar='<output>', help='Save generated dataset here.')
    parser.add_argument('-N', type=int, metavar='<N>', default=1000, help='No. of sample instances to generate.')
    parser.add_argument('--labels', type=str, nargs='+', default=['apple', 'banana', 'cherry'], metavar='<label>', help='No. of sample instances to generate.')
    A = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s', level=logging.INFO)

    label_combis = sum((list(combinations(A.labels, r)) for r in range(len(A.labels) + 1)), [])

    for i in range(A.N):
        _id = str(uuid4())
        A.output.write(json.dumps(dict(content=_id, id=_id, labels=choice(label_combis)), sort_keys=True))
        A.output.write('\n')
    #end for
    logger.info('{} samples generated in <{}>.'.format(A.N, A.output.name))
#end def


if __name__ == '__main__': main()



from collections import Counter, defaultdict
from datetime import datetime
import logging
import os
import pickle as pickle
from uuid import uuid4

import numpy as np
from numpy.lib.npyio import NpzFile

from scipy.sparse import csr_matrix, issparse

from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline

from tabulate import tabulate

from ..utils import Timer

__all__ = ['BaseFeaturizer', 'load_featurized', 'save_featurized', 'load_featurizer']

logger = logging.getLogger(__name__)


class BaseFeaturizer(Pipeline):
    def __init__(self, n_jobs=1, **kwargs):
        super(BaseFeaturizer, self).__init__(**kwargs)
        self.n_jobs = n_jobs
    #end def

    def fit(self, X, Y=None, **kwargs):
        self.uuid_ = str(uuid4())
        logger.debug('{} UUID is {}.'.format(type(self).__name__, self.uuid_))

        timer = Timer()
        super(BaseFeaturizer, self).fit(X, Y, **kwargs)
        logger.info('Done fitting {} {}.'.format(type(self).__name__, timer))

        self._post_fit(X=X, Y=Y, **kwargs)

        return self
    #end def

    def fit_transform(self, X, Y=None, **kwargs):
        self.uuid_ = str(uuid4())
        logger.debug('{} UUID is {}.'.format(type(self).__name__, self.uuid_))

        timer = Timer()
        X_featurized = super(BaseFeaturizer, self).fit_transform(X, Y, **kwargs)
        logger.info('Done fitting {} {}.'.format(type(self).__name__, timer))

        self._post_fit(X=X, X_featurized=X_featurized, Y=Y, **kwargs)

        return X_featurized
    #end def

    def save(self, f):
        if not hasattr(self, 'uuid_'):
            raise NotFittedError('This featurizer is not fitted yet.')

        pickle.dump(self, f, protocol=4)
        f.close()
        logger.info('Saved {} to <{}>.'.format(self, f.name))
    #end def

    def __str__(self):
        return '{}(uuid={})'.format(type(self).__name__, self.uuid_ if hasattr(self, 'uuid_') else 'None')
    #end def

    @property
    def uuid(self):
        if not hasattr(self, 'uuid_'):
            raise NotFittedError('This featurizer is not fitted yet.')

        return self.uuid_
    #end def

    def find(self, path):
        """Make it easier to find deep within a pipeline using filesystem like paths, i.e., `self.find('all_context_features/2/listcountvectorizer')`."""

        components = path.split('/')
        cur = self
        for name in components:
            if hasattr(cur, 'named_steps'):
                try:
                    i = int(name)
                    cur = cur.steps[i][1]
                except ValueError:
                    cur = cur.named_steps[name]

            elif hasattr(cur, 'transformer_list'):
                try:
                    i = int(name)
                    cur = cur.transformer_list[i][1]
                except ValueError:
                    for i in range(len(cur.transformer_list)):
                        if cur.transformer_list[i][0] == name:
                            cur = cur.transformer_list[i][1]
                            break
                        #end if
                    #end for

                    raise ValueError('Cannot find transformer with name "{}" in {}.'.format(name, cur))
                #end try

            else: raise TypeError('Unknown featurizer component type {}.'.format(repr(cur)))
        #end for

        return cur
    #end def

    def _post_fit(self, **kwargs): return self
#end def


def load_featurizer(f):
    featurizer = pickle.load(f)
    logger.info('Loaded {} from <{}>.'.format(featurizer, f.name))
    if not isinstance(featurizer, BaseFeaturizer):
        logger.warning('{} is not an instance of BaseFeaturizer. Perhaps you are loading the wrong file?'.format(featurizer))

    return featurizer
#end def


def save_featurized(f, X_featurized, *, Y_labels=None, featurized_at=datetime.utcnow(), **kwargs):
    _, ext = os.path.splitext(f.name)
    if ext.lower() != '.npz':
        logger.warning('Featurized files are Numpy compressed files and should have the .npz extension. It will definitely not work with the .gz extension.')

    if issparse(X_featurized):
        np.savez_compressed(f, X_featurized_data=X_featurized.data, X_featurized_indices=X_featurized.indices, X_featurized_indptr=X_featurized.indptr, X_featurized_shape=X_featurized.shape, Y_labels=Y_labels, featurized_at=featurized_at, **kwargs)  # Always use compression
    else:
        np.savez_compressed(f, X_featurized=X_featurized, Y_labels=Y_labels, featurized_at=featurized_at, **kwargs)  # Always use compression
    #end if
    f.close()

    logger.info('Saved {} featurized instances and its metadata to <{}>.'.format(X_featurized.shape[0], f.name))
#end def


def load_featurized(f, keys=[], raise_on_missing=True):
    timer = Timer()

    if f is None:
        logger.warning('No featurized file is specified. Will return empty instances.')
        o = defaultdict(None)
        o['X_featurized'] = np.empty((0, 1))
        o['Y_labels'] = np.empty((0, 1))
        f = type('test', (), {'name': 'file not specified'})()  # Hack to get around displaying empty file name
    else:
        o = np.load(f)
        if not isinstance(o, NpzFile):
            logger.warning('<{}> is not an instance of NpzFile. Perhaps you are loading the wrong file?'.format(f.name))
    #end if

    if not keys or 'X_featurized' in keys:
        if 'X_featurized_data' in o: X_featurized = csr_matrix((o['X_featurized_data'], o['X_featurized_indices'], o['X_featurized_indptr']), shape=o['X_featurized_shape'])
        else: X_featurized = o['X_featurized']

        logger.info('Loaded {} featurized instances from <{}> {}.'.format(X_featurized.shape[0], f.name, timer))
    #end if

    if keys and raise_on_missing:
        for k in keys:
            if k not in o and k not in ['X_featurized']:
                raise ValueError('{} not found in <{}>.'.format(k, f.name))
    #end if

    Y_labels = None
    if keys:
        d = tuple([X_featurized if k == 'X_featurized' else o[k] for k in keys])
        if 'Y_labels' in keys: Y_labels = d[keys.index('Y_labels')]
    else:
        d = dict((k, o[k]) for k in o.keys() if k not in {'X_featurized_data', 'X_featurized_indices', 'X_featurized_shape', 'X_featurized_indptr'})
        Y_labels = d.get('Y_labels')
    #end if

    if Y_labels is not None:
        try: freq = Counter(label for labels in Y_labels for label in labels)
        except TypeError: freq = Counter()
        freq['<none>'] = sum(1 for labels in Y_labels if not labels)
        if freq['<none>'] == 0: del freq['<none>']

        logger.info('Label frequencies for <{}>:\n{}'.format(f.name, tabulate(freq.most_common() + [('Labels total', sum(freq.values())), ('Cases total', len(Y_labels))], headers=('Label', 'Freq'), tablefmt='psql')))
    #end for

    if isinstance(d, tuple): return d
    return X_featurized, d
#end def



__all__ = ['HelloWorldFeaturizer']

import numpy as np

from ycml.featurizers import BaseFeaturizer
from ycml.transformers import PureTransformer


class RandomFeatures(PureTransformer):
    def __init__(self, feature_dims=10, **kwargs):
        self.feature_dims = feature_dims

    def transform(self, X):
        return np.random.rand(len(X), self.feature_dims)
#end class


class HelloWorldFeaturizer(BaseFeaturizer):
    def __init__(self, feature_dims=10, **kwargs):
        super(HelloWorldFeaturizer, self).__init__(steps=[('random_features', RandomFeatures(feature_dims=feature_dims))], **kwargs)

        self.feature_dims = feature_dims
    #end def
#end class



"""
This module provides wrapper function for transparently handling files regardless of location (local, cloud, etc).
"""
__all__ = ['uri_open', 'uri_to_tempfile', 'uri_read', 'uri_dump', 'uri_exists', 'get_uri_metadata', 'uri_exists_wait', 'URIFileType', 'URIType']

import functools
import warnings

import uriutils

warnings.simplefilter('default', category=DeprecationWarning)


def _deprecated(new_func):
    @functools.wraps(new_func)
    def f(*args, **kwargs):
        warnings.warn('The use of uriutils module in ycml.utils is deprecated. Please use the uriutils module instead.', category=DeprecationWarning, stacklevel=1)

        return new_func(*args, **kwargs)
    #end def

    return f
#end def


uri_open = _deprecated(uriutils.uri_open)
uri_read = _deprecated(uriutils.uri_read)
uri_dump = _deprecated(uriutils.uri_dump)
uri_exists = _deprecated(uriutils.uri_exists)
get_uri_metadata = _deprecated(uriutils.get_uri_metadata)
uri_exists_wait = _deprecated(uriutils.uri_exists_wait)


def uri_to_tempfile(uri, *, delete=True, **kwargs):
    warnings.warn('The use of uriutils methods in ycml.utils is deprecated. Please use the uriutils module instead.', category=DeprecationWarning, stacklevel=3)
    return uri_open(uri, in_memory=False, delete_tempfile=delete, **kwargs)
#end def


class URIFileType(uriutils.URIFileType):
    def __init__(self, *args, **kwargs):
        warnings.warn('The use of uriutils methods in ycml.utils is deprecated. Please use the uriutils module instead.', category=DeprecationWarning, stacklevel=1)
        super(URIFileType, self).__init__(*args, **kwargs)
    #end def
#end class


class URIType(uriutils.URIType):
    def __init__(self, *args, **kwargs):
        warnings.warn('The use of uriutils methods in ycml.utils is deprecated. Please use the uriutils module instead.', category=DeprecationWarning, stacklevel=1)
        super(URIType, self).__init__(*args, **kwargs)
    #end def
#end class



__all__ = ['ThresholdRescaler', 'get_thresholds_from_file']

import numpy as np

from ..utils import load_dictionary_from_file


class ThresholdRescaler(object):
    def __init__(self, thresholds, n_classes=None):
        if isinstance(thresholds, float): self.thresholds = np.full(n_classes, thresholds)
        elif isinstance(thresholds, (list, tuple)): self.thresholds = np.array(thresholds)
        else: self.thresholds = thresholds

        if self.thresholds.ndim == 2:
            self.thresholds = self.thresholds[:, 0]

        if self.thresholds.ndim != 1: raise ValueError('`thresholds` should be a 1D array.')

        if n_classes is None: n_classes = self.thresholds.shape[0]
        else: assert n_classes == self.thresholds.shape[0]

        self.denominators = 2.0 * (1.0 - self.thresholds)
    #end def

    def rescale(self, Y_proba):
        return rescale_proba_with_thresholds(Y_proba, self.thresholds, denominators=self.denominators)
    #end def

    def predict(self, Y_proba):
        Y_predict = np.zeros(Y_proba.shape, dtype=np.bool)
        for i in range(Y_proba.shape[0]):
            Y_predict[i, :] = Y_proba[i, :] > self.thresholds

        return Y_predict
    #end def
#end class


def rescale_proba_with_thresholds(Y_proba, thresholds, *, denominators=None):
    assert Y_proba.shape[1] == thresholds.shape[0]

    if denominators is None: denominators = 2.0 * (1.0 - thresholds)

    rescaled = np.zeros(Y_proba.shape)
    for i in range(Y_proba.shape[0]):
        for k in range(Y_proba.shape[1]):
            if Y_proba[i, k] >= thresholds[k]: rescaled[i, k] = 0.5 + ((Y_proba[i, k] - thresholds[k]) / denominators[k])
            else: rescaled[i, k] = Y_proba[i, k] / (thresholds[k] * 2.0)
        #end for
    #end for

    return rescaled
#end def


def get_thresholds_from_file(f, classes, *, default=0.5):
    d = load_dictionary_from_file(f)
    return np.array([float(d.get(c, default)) for c in classes])
#end def



__all__ = ['BaseFeatClass', 'load_featclass']

from argparse import ArgumentParser
import logging

import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin

from ..classifiers import load_classifier
from ..featurizers import load_featurizer

from uriutils import uri_open

from ..utils import get_settings, load_dictionary_from_file, get_class_from_module_path, chunked_iterator

logger = logging.getLogger(__name__)


class BaseFeatClass(BaseEstimator, ClassifierMixin):
    def __init__(self, featurizer_uri, classifier_uri, *, transform_args={}, predict_args={}, **kwargs):
        super(BaseFeatClass, self).__init__(**kwargs)

        self.featurizer_uri = featurizer_uri
        self.classifier_uri = classifier_uri

        self.transform_args = transform_args
        self.predict_args = predict_args

        with uri_open(featurizer_uri, mode='rb') as f:
            self.featurizer = load_featurizer(f)

        with uri_open(classifier_uri, mode='rb') as f:
            self.classifier = load_classifier(f)

        if 'featurizer_uuid' in kwargs and self.featurizer.uuid != kwargs['featurizer_uuid']:
            raise TypeError('Featurizer UUID mismatch: {} (this) != {} (<{}>)'.format(kwargs['featurizer_uuid'], self.featurizer.uuid, featurizer_uri))
        elif 'classifier_uuid' in kwargs and self.classifier.uuid != kwargs['classifier_uuid']:
            raise TypeError('Classifier UUID mismatch: {} (this) != {} (<{}>)'.format(kwargs['classifier_uuid'], self.classifier.uuid, classifier_uri))
    #end def

    def fit(self, *args, **kwargs):
        raise NotImplementedError('BaseFeatClass does not support the `fit` method.')

    def transform(self, X, **kwargs):
        return self.predict_proba(X, **kwargs)
    #end def

    def predict(self, X, **kwargs):
        predict_args = self.predict_args.copy()
        predict_args.update(kwargs)

        X_featurized = self.featurizer.transform(X, **self.transform_args)

        return self.classifier.predict(X_featurized, **predict_args)
    #end def

    def predict_proba(self, X, **kwargs):
        predict_args = self.predict_args.copy()
        predict_args.update(kwargs)

        X_featurized = self.featurizer.transform(X, **self.transform_args)

        return self.classifier.predict_proba(X_featurized, **predict_args)
    #end def

    def predict_and_proba(self, X, **kwargs):
        predict_args = self.predict_args.copy()
        predict_args.update(kwargs)

        X_featurized = self.featurizer.transform(X, **self.transform_args)

        return self.classifier.predict_and_proba(X_featurized, **predict_args)
    #end def

    def predictions_generator(self, instances_generator, *, chunk_size=100000, include_proba=True, unbinarized=True):
        for chunk in chunked_iterator(instances_generator, chunk_size):
            X = np.array(chunk, dtype=np.object)

            if include_proba:
                Y_proba, Y_predict = self.predict_and_proba(X)
                Y_proba_dicts = self.classifier.unbinarize_labels(Y_proba, to_dict=True)
            else:
                Y_predict = self.predict(X)
                Y_proba_dicts = None
            #end if

            Y_predict_list = self.classifier.unbinarize_labels(Y_predict, to_dict=False)

            if include_proba:
                if unbinarized:
                    yield from ((X[i], Y_predict_list[i], Y_proba_dicts[i]) for i in range(X.shape[0]))
                else:
                    yield from ((X[i], Y_predict[i], Y_proba[i]) for i in range(X.shape[0]))
            else:
                if unbinarized:
                    yield from ((X[i], Y_predict_list[i]) for i in range(X.shape[0]))
                else:
                    yield from ((X[i], Y_predict[i]) for i in range(X.shape[0]))
            #end if
        #end for
    #end def

    def decision_function(self, X, **kwargs):
        return self.predict_proba(X, **kwargs)

    def __str__(self):
        return 'BaseFeatClass(featurizer_uri={}, classifier_uri={})'.format(self.featurizer_uri, self.classifier_uri)
    #end def
#end class


def load_featclass(*, settings={}, uri=None, check_environment=True):
    settings_from_uri = {}

    if uri is not None:
        settings_from_uri = load_dictionary_from_file(uri)

    sources = ('env', settings_from_uri, settings) if check_environment else (settings_from_uri, settings)
    featclass_type = get_settings(key='featclass_type', sources=sources, raise_on_missing=True)
    featclass_parameters = get_settings(key='featclass_parameters', sources=sources, default={})

    featclass_class = get_class_from_module_path(featclass_type)
    featclass = featclass_class(**featclass_parameters)

    logger.info('Loaded {} from <{}>.'.format(featclass, uri if uri else ('env/settings' if check_environment else 'settings')))

    return featclass
#end def


def main():
    parser = ArgumentParser(description='Loads a featurizer-classifier. (DEVELOPMENT SCRIPT)')
    parser.add_argument('featclass_uri', type=str, metavar='<featclass_uri>', help='Featclass settings file to load.')
    A = parser.parse_args()

    logging.basicConfig(format=u'%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s', level=logging.INFO)

    featclass = load_featclass(uri=A.featclass_uri)
    print(featclass.transform(np.array([dict(content='this is a tort case. responsibility')])))
    print(featclass.predict_proba(np.array([dict(content='this is a tort case. responsibility')])))
#end def


if __name__ == '__main__': main()



__all__ = ['BaseClassifier', 'load_classifier']

from io import BytesIO
from datetime import datetime
import logging
import pickle
import tarfile
import time
from uuid import uuid4

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import NotFittedError

from ..utils import Timer
from ..utils import parse_n_jobs

logger = logging.getLogger(__name__)


class BaseClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, n_jobs=1, **kwargs):
        super(BaseClassifier, self).__init__()

        self.n_jobs = parse_n_jobs(n_jobs)
        self.n_jobs_string = n_jobs
    #end def

    def fit(self, X, Y, validation_data=None, **kwargs):
        self.uuid_ = str(uuid4())
        logger.debug('{} UUID is {}.'.format(self.name, self.uuid_))

        self.fitted_at_ = datetime.utcnow()

        timer = Timer()
        self._fit(X, Y, validation_data=validation_data, **kwargs)
        logger.info('{} fitting on {} instances complete {}.'.format(self.name, X.shape[0], timer))

        return self
    #end def

    def _fit(self, *args, **kwargs): raise NotImplementedError('_fit is not implemented.')

    def _predict_proba(self, X_featurized, **kwargs):
        raise NotImplementedError('_predict_proba is not implemented.')

    def predict_proba(self, X_featurized, **kwargs):
        timer = Timer()
        Y_proba = self._predict_proba(X_featurized, **kwargs)
        logger.debug('Computed prediction probabilities on {} instances {}.'.format(X_featurized.shape[0], timer))

        return Y_proba
    #end def

    def predict(self, X_featurized, **kwargs):
        timer = Timer()
        Y_proba = self._predict_proba(X_featurized, **kwargs)
        Y_predict = Y_proba >= 0.5
        logger.debug('Computed predictions on {} instances {}.'.format(X_featurized.shape[0], timer))

        return Y_predict
    #end def

    def predict_and_proba(self, X_featurized, **kwargs):
        timer = Timer()
        Y_proba = self._predict_proba(X_featurized, **kwargs)
        Y_predict = Y_proba >= 0.5
        logger.debug('Computed predictions and probabilities on {} instances {}.'.format(X_featurized.shape[0], timer))

        return Y_proba, Y_predict
    #end def

    def decision_function(self, *args, **kwargs): return self.predict_proba(*args, **kwargs)

    def save(self, f):
        if not hasattr(self, 'uuid_'):
            raise NotFittedError('This featurizer is not fitted yet.')

        with tarfile.open(fileobj=f, mode='w') as tf:
            with BytesIO() as model_f:
                try: pickle.dump(self, model_f, protocol=4)
                except pickle.PicklingError:
                    logger.error('PicklingError: Did you check to make sure that the classifier mixins (i.e., KerasNNClassifierMixin) is ahead of BaseClassifier in the MRO?')
                    raise
                #end try

                model_data = model_f.getvalue()
                model_f.seek(0)
                model_tarinfo = tarfile.TarInfo(name='model.pkl')
                model_tarinfo.size = len(model_data)
                model_tarinfo.mtime = int(time.time())
                tf.addfile(tarinfo=model_tarinfo, fileobj=model_f)
            #end with

            self.save_to_tarfile(tf)
        #end with

        f.close()

        logger.info('{} saved to <{}>.'.format(self, f.name))

        return self
    #end def

    def save_to_tarfile(self, tf): return self

    def load_from_tarfile(self, tf): return self

    @property
    def uuid(self):
        return self.uuid_
    #end def

    @property
    def name(self):
        return type(self).__name__

    @property
    def classes_(self): raise NotImplementedError('classes_ is not implemented.')

    def __str__(self):
        return '{}(UUID={})'.format(self.name, self.uuid_ if hasattr(self, 'uuid_') else 'None')
#end class


def load_classifier(f):
    with tarfile.open(fileobj=f, mode='r') as tf:
        classifier = pickle.load(tf.extractfile('model.pkl'))
        classifier.load_from_tarfile(tf)
    #end with

    logger.info('Loaded {} from <{}>.'.format(classifier, f.name))

    return classifier
#end def



__all__ = ['HelloWorldBinaryLabelsClassifier', 'HelloWorldMultiLabelsClassifier']

import numpy as np

from sklearn.svm import SVC

from ycml.classifiers import BinaryLabelsClassifier, MultiLabelsClassifier, MulticlassLabelsClassifier


class HelloWorldBinaryLabelsClassifier(BinaryLabelsClassifier):
    def fit_binarized(self, X_featurized, Y_binarized, **kwargs):
        self.classifier_ = SVC(probability=True).fit(X_featurized, Y_binarized[:, 1])

        return self
    #end def

    def _predict_proba(self, X_featurized, **kwargs):
        return self.classifier_.predict_proba(X_featurized)
#end class


class HelloWorldMultiLabelsClassifier(MultiLabelsClassifier):
    def fit_binarized(self, X_featurized, Y_binarized, **kwargs):
        self.classifiers_ = {}
        for j, c in enumerate(self.classes_):
            self.classifiers_[c] = SVC(probability=True).fit(X_featurized, Y_binarized[:, j])

        return self
    #end def

    def _predict_proba(self, X_featurized, **kwargs):
        Y_proba = np.zeros((X_featurized.shape[0], len(self.classes_)), dtype=np.float)

        for j, c in enumerate(self.classes_):
            Y_proba[:, j] = self.classifiers_[c].predict_proba(X_featurized, **kwargs)[:, 1]

        return Y_proba
    #end def
#end class


class HelloWorldMulticlassClassifier(MulticlassLabelsClassifier):
    def fit_binarized(self, X_featurized, Y_binarized, **kwargs):
        self.classifier_ = SVC(probability=True).fit(X_featurized, Y_binarized)

        return self
    #end def

    def _predict_proba(self, X_featurized, **kwargs):
        return self.classifier_.predict_proba(X_featurized, **kwargs)
        # Y_proba = np.zeros((X_featurized.shape[0], len(self.classes_)), dtype=np.float)

        # for j, c in enumerate(self.classes_):
        #     Y_proba[:, j] = self.classifiers_[c].predict_proba(X_featurized, **kwargs)[:, 1]

        # return Y_proba
    #end def
#end class



from argparse import ArgumentParser
import logging
import re

import numpy as np

from sklearn.metrics import classification_report as sklearn_classification_report

from tabulate import tabulate

from uriutils import URIFileType, URIType

from ..classifiers import load_classifier, get_thresholds_from_file
from ..featurizers import load_featurized
from ..utils import classification_report, find_best_thresholds, generate_pr_curves
from ..utils import get_settings
from ..utils import load_dictionary_from_file, save_dictionary_to_file

__all__ = []

logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser(description='Evaluate and analyze the performance of a trained classifier.')
    parser.add_argument('--log-level', type=str, metavar='<log_level>', help='Set log level of logger.')
    parser.add_argument('-s', '--settings', type=URIFileType(), metavar='<settings_file>', help='Settings file to configure models.')
    parser.add_argument('--n-jobs', type=int, metavar='<N>', help='No. of processor cores to use.')

    parser.add_argument('--classifier-info', type=URIFileType(), metavar='<classifier_file>', help='Display information about classifier.')

    parser.add_argument('-c', '--classifier', type=URIFileType(), metavar='<classifier_file>', help='Classifier file to use for evaluation.')
    parser.add_argument('-f', '--featurized', type=URIFileType(), metavar='<featurized_file>', nargs='+', help='Featurized instances for evaluation.')
    parser.add_argument('-t', '--thresholds', type=URIFileType(), metavar='<thresholds>', help='Threshold file to use for prediction.')

    parser.add_argument('--load-probabilities', type=URIFileType('rb'), metavar='<probabilities_file>', help='Load probabilities from here instead of recalculating.')
    parser.add_argument('--save-probabilities', type=URIFileType('wb'), metavar='<probabilities_file>', help='Save evaluation probabilities; useful for calibration.')

    parser.add_argument('--min-precision', type=float, metavar='<precision>', default=None, help='Set the minimum precision threshold.')
    parser.add_argument('--best-thresholds', type=URIFileType('w'), metavar='<thresholds_file>', help='Save best F1 threshold values here.')
    parser.add_argument('--minprec-thresholds', type=URIFileType('w'), metavar='<thresholds_file>', help='Save minimum precision best F1 threshold values here.')
    parser.add_argument('--clip-thresholds', type=float, nargs=2, metavar=('lower', 'upper'), default=None, help='Clip threshold values to given range. Only applied when saving to file.')
    parser.add_argument('--pr-curves', type=URIType(), metavar='<folder>', help='Save precision-recall curves in this folder.')

    A = parser.parse_args()

    file_settings = load_dictionary_from_file(A.settings) if A.settings else {}

    log_level = get_settings(key='log_level', sources=(A, 'env', file_settings), default='DEBUG').upper()
    log_format = get_settings(key='log_format', sources=(A, 'env', file_settings), default='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s')
    logging.basicConfig(format=log_format, level=logging.getLevelName(log_level))

    if A.classifier_info:
        classifier = load_classifier(A.classifier_info)
        params = classifier.get_params(deep=True)

        if logger.getEffectiveLevel() > logging.INFO: logger.error('Log level should be at least INFO to view classifier info.')

        tabulate_rows = [
            ('Type', type(classifier).__name__),
            ('UUID', classifier.uuid_),
            ('Fitted at', classifier.fitted_at_),
            ('Output labels', ', '.join(classifier.classes_)),
        ]
        tabulate_rows += [('Model Parameters', '{} = {}'.format(k, v)) for k, v in sorted(params.items())]
        tabulate_rows += [('Fitted Parameters', '{} = {}'.format(k, getattr(classifier, k))) for k in sorted(dir(classifier)) if re.match(r'.+[a-z]_$', k, flags=re.I) and k not in ['fitted_at_', 'uuid_', 'classes_', 'input_dims_', 'output_dims_']]
        logger.info('Classifier info for <{}>:\n{}'.format(A.classifier_file.name, tabulate(tabulate_rows, headers=('Key', 'Value'), tablefmt='psql')))
    #end if

    thresholds = None

    if A.load_probabilities:
        o = np.load(A.load_probabilities)
        logger.info('Loaded probabalities from <{}>.'.format(A.load_probabilities.name))

        Y_proba = o['Y_proba']
        Y_true_binarized = o['Y_true_binarized']
        thresholds = o['thresholds']
        if thresholds.dtype == np.object: thresholds = None
        labels = o['labels']

        if A.thresholds: thresholds = get_thresholds_from_file(A.thresholds, labels)
        classifier = 'Classifier({})'.format(o['classifier_uuid'])

    else:
        classifier = load_classifier(A.classifier)
        labels = classifier.classes_
        if A.thresholds: thresholds = get_thresholds_from_file(A.thresholds, labels)

        featurizer_uuid = None
        X_featurized, Y_labels = None, None
        for f in A.featurized:
            X, Y, uuid = load_featurized(f, keys=('X_featurized', 'Y_labels', 'featurizer_uuid'))
            if featurizer_uuid is None:
                featurizer_uuid = uuid
                X_featurized = X
                Y_labels = Y
            else:
                if uuid != featurizer_uuid: raise ValueError('<{}> has featurizer UUID "{}" which is inconsistent with <{}> UUID "{}".'.format(f.name, uuid, A.featurized[0].name, featurizer_uuid))
                X_featurized = np.concatenate((X_featurized, X), axis=0)
                Y_labels = np.concatenate((Y_labels, Y), axis=0)
            #end if
        #end for
        # X_featurized, Y_labels, featurizer_uuid = load_featurized(A.featurized, keys=('X_featurized', 'Y_labels', 'featurizer_uuid'))
        Y_proba, _ = classifier.predict_and_proba(X_featurized, thresholds=thresholds, binarized=True)
        Y_true_binarized = classifier.binarize_labels(Y_labels)

        N = X_featurized.shape[0]
        assert Y_true_binarized.shape[0] == N
    #end if

    logger.info('Classification report for <{}>:\n{}'.format(
        str(classifier),
        classification_report(Y_true_binarized, Y_proba, target_names=labels, thresholds=thresholds, precision_thresholds=A.min_precision, order='support'))
    )

    # logger.info('sklearn Classification report for <{}>:\n{}'.format(
    #     str(classifier),
    #     sklearn_classification_report(Y_true_binarized, Y_proba, target_names=labels, digits=3))
    # )

    if A.save_probabilities:
        np.savez_compressed(A.save_probabilities, featurizer_uuid=featurizer_uuid, classifier_uuid=classifier.uuid_, Y_proba=Y_proba, Y_true_binarized=Y_true_binarized, thresholds=thresholds, labels=labels)
        logger.info('Saved evaluation probabilities to <{}>.'.format(A.save_probabilities.name))
    #end if

    if A.best_thresholds:
        best_thresholds = find_best_thresholds(Y_true_binarized, Y_proba, target_names=labels)
        if A.clip_thresholds:
            best_thresholds = best_thresholds.clip(A.clip_thresholds[0], A.clip_thresholds[1])
            logger.info('Best threshold values clipped to [{}, {}].'.format(*A.clip_thresholds))
        #end if

        o = dict((c, float(best_thresholds[i])) for i, c in enumerate(labels))
        save_dictionary_to_file(A.best_thresholds, o, title='thresholds')
    #end if

    if A.minprec_thresholds:
        minprec_thresholds = find_best_thresholds(Y_true_binarized, Y_proba, precision_thresholds=A.min_precision, target_names=labels)
        if A.clip_thresholds:
            minprec_thresholds = minprec_thresholds.clip(A.clip_thresholds[0], A.clip_thresholds[1])
            logger.info('Min-precision threshold values clipped to [{}, {}].'.format(*A.clip_thresholds))
        #end if

        o = dict((c, float(minprec_thresholds[i])) for i, c in enumerate(labels))
        save_dictionary_to_file(A.minprec_thresholds, o, title='thresholds')
    #end if

    if A.pr_curves:
        generate_pr_curves(Y_true_binarized, Y_proba, A.pr_curves.geturl(), target_names=labels, thresholds=thresholds, precision_thresholds=A.min_precision)
    #end if
#end def


if __name__ == '__main__': main()



__all__ = ['load_dictionary_from_file', 'save_dictionary_to_file', 'get_settings', 'bulk_get_settings']

import json
import logging
import os
import pickle

try: import yaml
except ImportError: pass

from uriutils import uri_open

logger = logging.getLogger(__name__)


def load_dictionary_from_file(file_or_filename, *, force_format=None, title='dictionary'):
    if isinstance(file_or_filename, str): file_or_filename = uri_open(file_or_filename, 'rb')

    if force_format: ext = '.' + force_format
    else: _, ext = os.path.splitext(file_or_filename.name.lower())

    if ext == '.json': o = json.load(file_or_filename)
    elif ext == '.yaml': o = yaml.load(file_or_filename)
    elif ext == '.pickle': o = pickle.load(file_or_filename)
    else: raise ValueError('<{}> is an unrecognized format for dictionary. Only JSON and YAML are supported right now.')

    logger.info('Loaded {} {} from <{}>.'.format(ext[1:].upper(), title, file_or_filename.name))

    return o
#end def


def save_dictionary_to_file(f, d, *, force_format=None, title='dictionary', **kwargs):
    if isinstance(f, str): f = open(f, 'w')

    if force_format: ext = '.' + force_format
    else: _, ext = os.path.splitext(f.name.lower())

    if ext == '.json':
        kwargs.setdefault('indent', 4)
        kwargs.setdefault('sort_keys', True)
        json.dump(d, f, **kwargs)

    elif ext == '.yaml':
        kwargs.setdefault('default_flow_style', False)
        yaml.dump(d, f, **kwargs)

    elif ext == '.pickle':
        pickle.dump(d, f, **kwargs)

    else: raise ValueError('<{}> is an unrecognized format for {}. Only JSON and YAML are supported right now.'.format(title))

    logger.info('Saved {} {} to <{}>.'.format(ext[1:].upper(), title, f.name))
#end def


ENV_SOURCE_KEYS = ['env', 'environment', 'ENV']


def get_settings(*source_key_pairs, key=None, keys=None, sources=None, default=None, raise_on_missing=None, parse_string_func=None, auto_parse=False):
    source_key_pairs = list(source_key_pairs)
    if key and sources:
        source_key_pairs += [(src, key) for src in sources]

    if keys and sources:
        source_key_pairs += [(src, k) for src in sources for k in keys]

    for src, key in source_key_pairs:
        v = None
        for key_ in (key, key.upper() if not key.isupper() else None, key.lower() if not key.islower() else None):
            if key_ is None: continue

            if hasattr(src, key_): v = getattr(src, key_)
            elif hasattr(src, 'get'): v = src.get(key_)
            elif src in ENV_SOURCE_KEYS: v = os.environ.get(key_)

            if v is not None:
                if isinstance(v, str):
                    if parse_string_func: return parse_string_func(v)
                    elif auto_parse: return _auto_parse(v)
                #end if

                return v
            #end if
        #end for
    #end for

    if raise_on_missing:
        raise ValueError('Unable to find setting [{}].'.format(', '.join(sorted(set(key for _, key in source_key_pairs)))))

    return default
#end def


def _auto_parse(s):
    try: return int(s)
    except ValueError: pass

    try: return float(s)
    except ValueError: pass

    try: return json.loads(s)
    except json.JSONDecodeError: pass

    return s
#end def


def bulk_get_settings(*sources, normalize_func=None, auto_parse=False):
    bulk_settings = {}
    chosen_sources = {}

    def _normalize_key(k):  # settings key are always lowercased. IRregardless.
        k = k.lower()
        normalized = normalize_func(k)
        return normalized if normalized else k
    #end def

    for i, src in enumerate(sources):
        if src in ENV_SOURCE_KEYS:
            for k, v in os.environ.items():
                norm_k = _normalize_key(k)
                bulk_settings[norm_k] = v
                chosen_sources[norm_k] = 'env'
            #end for
        elif hasattr(src, 'items'):
            for k, v in src.items():
                norm_k = _normalize_key(k)
                bulk_settings[norm_k] = v
                chosen_sources[norm_k] = 'dict_{}'.format(i)
            #end for

        else:
            for k in filter(lambda a: not a.startswith('_'), dir(src)):
                v = getattr(src, k)
                if callable(v): continue  # skip functions, we only want values

                norm_k = _normalize_key(k)
                bulk_settings[norm_k] = v
                chosen_sources[norm_k] = 'obj_{}'.format(i)
            #end for
        #end if
    #end for

    if auto_parse:
        for k, v in bulk_settings.items():
            if isinstance(v, str):
                bulk_settings[k] = _auto_parse(v)
    #end if

    return bulk_settings, chosen_sources
#end def



__all__ = ['MultiLabelsFeatureSelector']

from .labels import MultiLabelsClassifier
from ..utils import get_class_from_module_path

import numpy as np


class MultiLabelsFeatureSelector(MultiLabelsClassifier):
    def __init__(self, score_func=None, **kwargs):
        super(MultiLabelsClassifier, self).__init__(**kwargs)

        self.score_func = get_class_from_module_path(score_func)
    #end def

    def fit_binarized(self, X_featurized, Y_binarized, **kwargs):
        scores = np.zeros((X_featurized.shape[1], Y_binarized.shape[1]))
        pval = np.zeros((X_featurized.shape[1], Y_binarized.shape[1]))
        for j in range(Y_binarized.shape[1]):
            scores[:, j], pval[:, j] = self.score_func(X_featurized, Y_binarized[:, j])

        self.labels_scores_ = scores
        self.labels_pvals_ = pval
        self.scores_ = np.max(scores, axis=1)

        return self
    #end def

    def feature_select(self, X_featurized, Y_labels, **kwargs):
        self.fit(X_featurized, Y_labels, **kwargs)

        return self.scores_
    #end def

    def transform(self, X, **kwargs):
        return X
#end class



__all__ = ['TokensToIndexTransformer']

import logging

from .base import PureTransformer
from .text import ListCountVectorizer

logger = logging.getLogger(__name__)


class TokensToIndexTransformer(PureTransformer):
    def __init__(self, skip_unknown=False, pad_sequences=None, count_vectorizer_args={}, pad_sequences_args={}, **kwargs):
        super(TokensToIndexTransformer, self).__init__(**kwargs)

        self.skip_unknown = skip_unknown
        self.pad_sequences = pad_sequences
        self.count_vectorizer_args = count_vectorizer_args
        self.pad_sequences_args = pad_sequences_args
    #end def

    def fit(self, X, *args, **kwargs):
        self.count_vectorizer_ = ListCountVectorizer(**self.count_vectorizer_args).fit(X)
        logger.debug('TokensToIndexTransformer vocabulary fitted with size {}.'.format(len(self.vocabulary_)))

        return self
    #end def

    def _transform(self, X, y=None, **kwargs):
        if 'maxlen' in self.pad_sequences_args:
            raise ValueError('The `maxlen` argument should not be set in `pad_sequences_args`. Set it in `pad_sequences` instead.')

        analyzer = self.count_vectorizer_.build_analyzer()
        V = self.vocabulary_

        X_transformed = []
        for seq in X:
            indexes = []
            for j, tok in enumerate(analyzer(seq)):
                index = V.get(tok)

                if not getattr(self, 'skip_unknown', False): indexes.append(0 if index is None else (index + 1))
                elif index is not None: indexes.append(index)
            #end for

            X_transformed.append(indexes)
        #end for

        if self.pad_sequences is not None:
            from keras.preprocessing.sequence import pad_sequences as keras_pad_sequences

            maxlen = getattr(self, 'pad_sequences_maxlen_', None if self.pad_sequences is True else self.pad_sequences)
            X_transformed = keras_pad_sequences(X_transformed, maxlen=maxlen, **self.pad_sequences_args)
            if self.pad_sequences is True or maxlen is not None:
                logger.debug('TokensToIndexTransformer transformed sequences has max length {}.'.format(X_transformed.shape[1]))

            self.pad_sequences_maxlen_ = X_transformed.shape[1]
        #end if

        return X_transformed
    #end def

    @property
    def vocabulary_(self): return self.count_vectorizer_.vocabulary_

    @property
    def stop_words_(self): return self.count_vectorizer_.stop_words_

    def __repr__(self):
        count_vectorizer_repr = '{}(vocabulary_={}, stop_words_={})'.format(self.count_vectorizer_.__class__.__name__, len(getattr(self.count_vectorizer_, 'vocabulary_', [])), len(getattr(self.count_vectorizer_, 'stop_words_', []))) if hasattr(self, 'count_vectorizer_') else None

        return '{}(skip_unknown={}, pad_sequences={}, count_vectorizer_args={}, pad_sequences_args={}, count_vectorizer_={})'.format(self.__class__.__name__, self.skip_unknown, self.pad_sequences, self.count_vectorizer_args, self.pad_sequences_args, count_vectorizer_repr)
    #end def
#end class



__all__ = ['check_api_token']

from functools import wraps

from flask import current_app, abort


def check_api_token(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        api_token = kwargs.pop('api_token')
        if api_token != current_app.config['api_token']: abort(403)

        return f(*args, **kwargs)
    #end def

    return decorated_function
#end def



__all__ = ['chunked_iterator']

from itertools import islice


def chunked_iterator(iterable, n):
    it = iter(iterable)
    while True:
       chunk = tuple(islice(it, n))
       if not chunk: return
       yield chunk
    #end while
#end def



__all__ = ['DictExtractionTransformer', 'DictListExtractionTransformer', 'FunctionTransformer', 'IdentityTransformer']

from . import PureTransformer


class DictExtractionTransformer(PureTransformer):
    """Extract a given key from dictionary object."""

    def __init__(self, key=None, default=None, **kwargs):
        kwargs.setdefault('nparray', False)
        super(DictExtractionTransformer, self).__init__(**kwargs)

        self.key = key
        self.default = default
    #end def

    def transform_one(self, d, **kwargs): return d.get(self.key, self.default)
#end class


class DictListExtractionTransformer(PureTransformer):
    """Extract a given key from list of dictionary object."""

    def __init__(self, key=None, default=None, **kwargs):
        super(DictListExtractionTransformer, self).__init__(**kwargs)

        self.key = key
        self.default = default
    #end def

    def transform_one(self, L, **kwargs): return [d.get(self.key, self.default) for d in L]
#end class


class FunctionTransformer(PureTransformer):
    def __init__(self, func, **kwargs):
        super(FunctionTransformer, self).__init__(**kwargs)
        self.func = func
    #end def

    def transform_one(self, x, **kwargs):
        return self.func(x)
#end class


class IdentityTransformer(PureTransformer):
    def transform_one(self, x, **kwargs):
        return x
#end class



__all__ = ['load_instances', 'shuffle_instances']

from argparse import ArgumentParser
from collections import Counter
import json
import logging

import numpy as np

from tabulate import tabulate

from uriutils import URIFileType

from .timer import Timer

logger = logging.getLogger(__name__)


def load_instances(instance_files, labels_field='labels', display_threshold=None, limit=None, progress_interval=None):
    if not isinstance(instance_files, (list, tuple)): instance_files = [instance_files]

    def _filter_display(d, threshold):
        if threshold is None:
            yield from d
        else:
            ignored_count, ignored_freq = 0, 0
            for label, freq in d:
                if freq < threshold:
                    ignored_count += 1
                    ignored_freq += freq
                else:
                    yield label, freq
                #end if
            #end for

            if ignored_count > 0:
                yield 'Ignored labels with freq < {}'.format(threshold), '{} ({} labels)'.format(ignored_freq, ignored_count)
        #end if
    #end def

    total_count = 0
    labels_freq = Counter()
    for f in instance_files:
        freq = Counter()
        count = 0
        timer = Timer()

        try:
            for lineno, line in enumerate(f, start=1):
                o = json.loads(line)

                if labels_field is None or labels_field is False:
                    yield o
                else:
                    labels = o.get(labels_field)
                    if labels is not None:
                        try:
                            if labels: freq.update(labels)
                            else: freq['<none>'] += 1
                        except TypeError: pass  # unhashable type
                    #end if

                    yield (o, labels)
                #end if

                count += 1
                total_count += 1
                if progress_interval and count % progress_interval == 0: logger.debug('{} instances read from <{}> so far.'.format(count, f.name))
                if limit and total_count == limit: break
            #end for
        except EOFError as e: logger.exception('Exception while reading instances. Will skip the remaining of <{}>.'.format(f.name))
        #end try

        logger.info('{} instances read from file <{}> {}.'.format(count, f.name, timer))

        if labels_field:
            labels_freq += freq
            threshold = freq.most_common()[99][1] if len(freq) > 100 and display_threshold is None else display_threshold
            logger.info('Label frequencies for <{}>:\n{}'.format(f.name, tabulate(list(_filter_display(freq.most_common(), threshold)) + [('Labels total', sum(freq.values())), ('Total', count)], headers=('Label', 'Freq'), tablefmt='psql')))
        #end if

        if limit and total_count >= limit: break
    #end for

    if labels_field and len(instance_files) > 1:
        threshold = freq.most_common()[99][1] if len(freq) > 100 and display_threshold is None else display_threshold
        logger.info('Total label frequencies:\n{}'.format(tabulate(list(_filter_display(labels_freq.most_common(), threshold)) + [('Labels total', sum(labels_freq.values())), ('Total', total_count)], headers=('Label', 'Freq'), tablefmt='psql')))
#end def


def shuffle_instances(X, Y, *, limit=None):
    indices = np.arange(X.shape[0])
    np.random.shuffle(indices)
    if limit: indices = indices[:limit]

    return X[indices], Y[indices]
#end def


def main():
    parser = ArgumentParser(description='Quick utility for getting instance information.')
    parser.add_argument('instances', type=URIFileType(mode='r'), nargs='+', metavar='<instances>', help='List of instance files to get information about.')
    parser.add_argument('-l', '--label-key', type=str, metavar='<key>', default='labels', help='The key name for the label.')
    A = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s', level=logging.INFO)

    list(load_instances(A.instances, labels_field=A.label_key))
#end def


if __name__ == '__main__': main()



__all__ = ['KerasNNClassifierMixin', 'keras_f1_score', 'EarlyStopping']

import logging
import math
import os
import shutil
from tempfile import NamedTemporaryFile

try:
    from keras import backend as K
    from keras.callbacks import Callback, ModelCheckpoint
    from keras.models import load_model

    import tensorflow as tf
except ImportError:
    Callback = object

import numpy as np
import scipy.sparse as sps

from sklearn.model_selection import train_test_split

from uriutils import uri_open

from ..utils import Timer

logger = logging.getLogger(__name__)


class KerasNNClassifierMixin(object):
    PICKLE_IGNORED_ATTRIBUTES = set()
    NN_MODEL_ATTRIBUTE = 'nn_model_'

    def __init__(
        self,
        tf_config=None, set_session=True,  # set_session should be False when you are initializing a second classifier...
        epochs=10, batch_size=128, passes_per_epoch=1,
        initial_weights=None, initial_epoch=0,
        validation_size=0.2, verbose=0,
        early_stopping=None, save_best=None, save_weights=None,
        log_device_placement=False,
        **kwargs
    ):
        self.tf_config = tf_config

        self.epochs = epochs
        self.batch_size = batch_size
        self.passes_per_epoch = passes_per_epoch

        self.initial_weights = initial_weights
        self.initial_epoch = initial_epoch

        self.validation_size = validation_size
        self.verbose = verbose

        self.early_stopping = early_stopping
        self.save_weights = save_weights
        self.save_best = save_best

        self.log_device_placement = log_device_placement

        if set_session: self.set_session(tf_config)
    #end def

    def fit_binarized(self, X_featurized, Y_binarized, **kwargs):
        setattr(self, self.NN_MODEL_ATTRIBUTE, self.compile_model(**kwargs))
        self.keras_fit_generator(X_featurized, Y_binarized, **kwargs)

        return self
    #end def

    def compile_model(self, X_featurized, Y_binarized, **kwargs): raise NotImplementedError('compile_model is not implemented.')

    def _predict_proba(self, X_featurized, batch_size=1024, use_generator=False, **kwargs):
        if X_featurized.shape[0] < batch_size: kwargs['verbose'] = 0

        if use_generator or X_featurized.shape[0] > batch_size: Y_proba = self.keras_predict_generator(X_featurized, batch_size=batch_size, **kwargs)
        else: Y_proba = self.keras_predict(X_featurized, batch_size=batch_size, **kwargs)

        return Y_proba
    #end def

    def set_session(self, tf_config=None):
        if tf_config is None:
            tf_config = self.tf_config

        if tf_config is None:
            n_jobs = getattr(self, 'n_jobs', 1)
            log_device_placement = getattr(self, 'log_device_placement', logger.getEffectiveLevel() <= logging.DEBUG)
            tf_config = tf.ConfigProto(inter_op_parallelism_threads=n_jobs, intra_op_parallelism_threads=n_jobs, log_device_placement=log_device_placement, allow_soft_placement=True)
        #end if

        tf_session = tf.Session(config=tf_config)
        K.set_session(tf_session)

        init_op = tf.global_variables_initializer()
        tf_session.run(init_op)
    #end def

    def keras_fit(self, X, Y, *, nn_model=None, validation_data=None, resume=None, **fit_args):
        if nn_model is None: nn_model = getattr(self, self.NN_MODEL_ATTRIBUTE)

        if not self._pre_fit_setup(nn_model, resume=resume, **fit_args): return

        if sps.issparse(X): X = X.toarray()
        if sps.issparse(Y): Y = Y.toarray()
        if validation_data is not None:
            X_validation, Y_validation = validation_data
            validation_data = (X_validation.toarray() if sps.issparse(X_validation) else X_validation, Y_validation.toarray() if sps.issparse(Y_validation) else Y_validation)
        #end if

        logger.info('{} instances used for training and {} instances used for validation.'.format(Y.shape[0], validation_data[1].shape[0] if validation_data else int(self.validation_size * Y.shape[0])))

        return nn_model.fit(X, Y, validation_data=validation_data, validation_split=0.0 if validation_data is not None else self.validation_size, epochs=self.epochs, batch_size=self.batch_size, verbose=self.verbose, callbacks=self.build_callbacks(), initial_epoch=self.initial_epoch, **fit_args)
    #end def

    def keras_fit_generator(self, X, Y, *, nn_model=None, generator_func=None, validation_data=None, resume=None, **fit_args):
        if nn_model is None: nn_model = getattr(self, self.NN_MODEL_ATTRIBUTE)

        if not self._pre_fit_setup(nn_model, resume=resume, **fit_args): return

        if validation_data is None:
            X_train, X_validation, Y_train, Y_validation = train_test_split(X, Y, test_size=self.validation_size)
            validation_data = (X_validation.toarray() if sps.issparse(X_validation) else X_validation, Y_validation.toarray() if sps.issparse(Y_validation) else Y_validation)
        else:
            X_train, Y_train = X, Y
            X_validation, Y_validation = validation_data
            validation_data = (X_validation.toarray() if sps.issparse(X_validation) else X_validation, Y_validation.toarray() if sps.issparse(Y_validation) else Y_validation)
        #end if
        N_train = X_train.shape[0]
        logger.info('{} instances used for training and {} instances used for validation.'.format(N_train, validation_data[1].shape[0]))

        steps_per_epoch = int((N_train * self.passes_per_epoch) / self.batch_size)
        if steps_per_epoch <= 0: raise ValueError('steps_per_epoch ({}) is <= 0!'.format(steps_per_epoch))
        logger.info('Fit generator will run {} steps per epoch with batch size of {}. This will make 1 pass through the training data in {:.2f} epochs.'.format(steps_per_epoch, self.batch_size, N_train / (steps_per_epoch * self.batch_size)))

        if generator_func is None: generator_func = self._generator

        return nn_model.fit_generator(
            generator_func(X_train, Y_train, batch_size=self.batch_size),
            validation_data=validation_data,
            initial_epoch=self.initial_epoch,
            steps_per_epoch=steps_per_epoch, epochs=self.epochs, verbose=self.verbose, callbacks=self.build_callbacks(),
            **fit_args
        )
    #end def

    def _pre_fit_setup(self, nn_model, *, resume=None, **kwargs):
        if resume:
            self.initial_weights = resume[0]

            if len(resume) > 1:
                self.initial_epoch = int(resume[1])
                if self.initial_epoch < 0:
                    logger.info('Detected negative initial epoch value. Will not perform model fitting.')
                    self.epochs = 0
                #end if
            #end if
        #end if

        if self.initial_weights:
            with uri_open(self.initial_weights, in_memory=False) as f:
                nn_model.load_weights(f.temp_name)
            logger.info('Loaded initial weights file from <{}> will start at epoch {}.'.format(self.initial_weights, self.initial_epoch))
        #end if

        if self.epochs == 0:
            logger.warning('Epochs is set to 0. Model fitting will not continue.')
            return
        #end if

        return True
    #end def

    def _generator(self, X, Y, *, batch_size=128, shuffle=True):
        N = X.shape[0]
        if batch_size > N: raise ValueError('batch_size ({}) is > than number of instances ({}).'.format(batch_size, N))

        sparse_X = sps.issparse(X)
        if Y is None: batch_sized_zeros = np.zeros(batch_size)

        if shuffle:
            shuffled_indexes = np.random.permutation(N)
            X_shuffled = X[shuffled_indexes, ...]
            if Y is not None: Y_shuffled = Y[shuffled_indexes]
        else: X_shuffled, Y_shuffled = X, Y

        cur = 0
        while True:
            if cur + batch_size > N:
                if shuffle:
                    shuffled_indexes = np.random.permutation(N)
                    X_shuffled = X[shuffled_indexes, :]
                    if Y is not None: Y_shuffled = Y[shuffled_indexes]
                #end if

                cur = 0
            #end if

            X_batch = X_shuffled[cur:cur + batch_size, ...].toarray() if sparse_X else X_shuffled[cur:cur + batch_size, ...]
            Y_batch = batch_sized_zeros if Y is None else Y_shuffled[cur:cur + batch_size]

            yield (X_batch, Y_batch)

            cur += batch_size
        #end while
    #end def

    def keras_predict(self, X, *, nn_model=None, **kwargs):
        if nn_model is None: nn_model = getattr(self, self.NN_MODEL_ATTRIBUTE)

        batch_size = kwargs.pop('batch_size', self.batch_size)
        verbose = kwargs.pop('verbose', self.verbose)

        return nn_model.predict_proba(X.toarray() if sps.issparse(X) else X, batch_size=batch_size, verbose=verbose)
    #end def

    def keras_predict_generator(self, X, *, nn_model=None, **kwargs):
        if nn_model is None: nn_model = getattr(self, self.NN_MODEL_ATTRIBUTE)

        N = X.shape[0]

        batch_size = kwargs.pop('batch_size', self.batch_size)
        verbose = kwargs.pop('verbose', self.verbose)
        steps = int(math.ceil(N / batch_size))

        return nn_model.predict_generator(self._generator(X, None, batch_size=batch_size, shuffle=False), steps=steps, verbose=verbose)[:N, ...]
    #end def

    def build_callbacks(self):
        callbacks = []
        if self.early_stopping is not None:
            early_stopping = EarlyStopping(**self.early_stopping)
            callbacks.append(early_stopping)
            logger.info('Set up {}.'.format(early_stopping))
        #end if

        if self.save_best is not None:
            monitor = self.save_best.get('monitor', 'accuracy')
            filepath = self.save_best.get('path', '{epoch:04d}_{' + monitor + ':.5f}.h5')
            callbacks.append(ModelCheckpoint(filepath=filepath, monitor=monitor, verbose=self.verbose, save_best_only=True, save_weights_only=True, mode='max'))
            logger.info('Best `{}` model will be saved to <{}>.'.format(monitor, filepath))
        #end if

        if self.save_weights is not None:
            path = self.save_weights.get('path', '.')
            period = self.save_weights.get('period', 1)
            filepath = os.path.join(self.save_weights, '{epoch:04d}_{val_keras_f1_score:.5f}.h5') if os.path.isdir(path) else path
            callbacks.append(ModelCheckpoint(filepath=filepath, verbose=self.verbose, save_best_only=False, save_weights_only=True, period=period))
            logger.info('Weights of every {}th model will be saved to <{}>.'.format(period, filepath))
        #end if

        return callbacks
    #end def

    def save_to_tarfile(self, tar_file):
        with NamedTemporaryFile(prefix=__name__ + '.', suffix='.h5', delete=False) as f:
            temp_path = f.name
        #end with

        self.nn_model_.save(temp_path)
        tar_file.add(temp_path, arcname='nn_model.h5')
        os.remove(temp_path)

        return self
    #end def

    def load_from_tarfile(self, tar_file):
        self.set_session()

        fname = None
        try:
            with NamedTemporaryFile(suffix='.h5', delete=False) as f:
                timer = Timer()
                shutil.copyfileobj(tar_file.extractfile('nn_model.h5'), f)
                fname = f.name
            #end with

            self.nn_model_ = load_model(fname, custom_objects=self.custom_objects)
            logger.debug('Loaded neural network model weights {}.'.format(timer))

        finally:
            if fname:
                os.remove(fname)
        #end try

        return self
    #end def

    def __getstate__(self):
        state = super(KerasNNClassifierMixin, self).__getstate__()

        ignored_attrs = set([self.NN_MODEL_ATTRIBUTE, 'tf_session']) | self.PICKLE_IGNORED_ATTRIBUTES
        for k in ignored_attrs:
            if k in state:
                del state[k]

        return state
    #end def

    @property
    def custom_objects(self): return {}
#end class


def keras_f1_score(y_true, y_pred):
    '''Calculates the F score, the weighted harmonic mean of precision and recall.
    This is useful for multi-label classification, where input samples can be
    classified as sets of labels. By only using accuracy (precision) a model
    would achieve a perfect score by simply assigning every class to every
    input. In order to avoid this, a metric should penalize incorrect class
    assignments as well (recall). The F-beta score (ranged from 0.0 to 1.0)
    computes this, as a weighted mean of the proportion of correct class
    assignments vs. the proportion of incorrect class assignments.
    With beta = 1, this is equivalent to a F-measure. With beta < 1, assigning
    correct classes becomes more important, and with beta > 1 the metric is
    instead weighted towards penalizing incorrect class assignments.
    '''
    # If there are no true positives, fix the F score at 0 like sklearn.

    if K.sum(K.round(K.clip(y_true, 0, 1))) == 0:
        return 0

    beta = 1.0

    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    p = true_positives / (predicted_positives + K.epsilon())

    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    r = true_positives / (possible_positives + K.epsilon())

    bb = beta ** 2
    fbeta_score = (1 + bb) * (p * r) / (bb * p + r + K.epsilon())

    return fbeta_score
#end def


class EarlyStopping(Callback):
    def __init__(self, monitor='val_accuracy', patience=5, min_delta=0.01, min_epoch=50):
        super(EarlyStopping, self).__init__()

        self.patience = patience
        self.min_delta = min_delta
        self.min_epoch = min_epoch
        self.monitor = monitor

        self.val_scores = []
    #end def

    def on_epoch_end(self, epoch, logs={}):
        monitor_score = logs[self.monitor]
        self.val_scores.append(monitor_score)
        if len(self.val_scores) < self.patience or epoch <= self.min_epoch or monitor_score < 0.2:  # hard limit
            return

        if self.min_epoch == epoch + 1:
            logger.info('Epoch {} > {} (min_epoch). Starting early stopping monitor.'.format(epoch, self.min_epoch))

        m = np.mean(self.val_scores[-self.patience - 1:-1])
        delta = abs(monitor_score - m)
        min_delta = self.min_delta * m
        # logger.debug('mean({}[-{}:])={}; delta={}; min_delta={};'.format(self.monitor, self.patience, m, delta, min_delta))

        if delta < min_delta:
            logger.info('Model delta fell below `min_delta` threshold. Early stopped.')
            self.model.stop_training = True
        #end if
    #end def

    def __str__(self):
        return 'EarlyStopping(monitor={}, patience={}, min_delta={}, min_epoch={})'.format(self.monitor, self.patience, self.min_delta, self.min_epoch)
#end def



__all__ = ['ThresholdingFeatClass']

import logging

import numpy as np

from .base import BaseFeatClass
from ..classifiers import ThresholdRescaler, BinaryLabelsClassifier
from ..classifiers import get_thresholds_from_file

logger = logging.getLogger(__name__)


class ThresholdingFeatClass(BaseFeatClass):
    def __init__(self, thresholds=None, thresholds_uri=None, **kwargs):
        super(ThresholdingFeatClass, self).__init__(**kwargs)

        if thresholds is not None and thresholds_uri is not None: raise ValueError('Only one of thresholds/thresholds_uri should be specified.')

        self.thresholds_uri = thresholds_uri

        if thresholds is None:
            if self.thresholds_uri:
                thresholds = get_thresholds_from_file(self.thresholds_uri, self.classifier.classes_)

                if len(self.classifier.classes_) == 2 and not np.isclose(thresholds.sum(), 1.0) and isinstance(self.classifier, BinaryLabelsClassifier):
                    if thresholds[0] == 0.5: thresholds[0] = 1.0 - thresholds[1]
                    if thresholds[1] == 0.5: thresholds[1] = 1.0 - thresholds[0]

                    logger.warning('Thresholds were set automatically for BinaryLabelsClassifier to {}={} and {}={}.'.format(self.classifier.classes_[0], thresholds[0], self.classifier.classes_[1], thresholds[1]))
                #end if

            else: thresholds = 0.5
        #end if

        self.rescaler = ThresholdRescaler(thresholds, len(self.classifier.classes_))
    #end def

    def predict(self, X, **kwargs):
        return self.rescaler.predict(super(ThresholdingFeatClass, self).predict_proba(X, **kwargs))
    #end def

    def predict_proba(self, X, *, rescale=True, **kwargs):
        Y_proba = super(ThresholdingFeatClass, self).predict_proba(X, **kwargs)

        if rescale:
            return self.rescaler.rescale(Y_proba)

        return Y_proba
    #end def

    def predict_and_proba(self, X, *, rescale=True, **kwargs):
        Y_proba = super(ThresholdingFeatClass, self).predict_proba(X, **kwargs)

        if rescale:
            Y_proba = self.rescaler.rescale(Y_proba)
            Y_predict = Y_proba >= 0.5
        else:
            Y_predict = self.rescaler.predict(Y_proba)
        #end if

        return Y_proba, Y_predict
    #end def

    def decision_function(self, X, **kwargs):
        return self.predict_proba(X, **kwargs)

    def __str__(self):
        return 'ThresholdingFeatClass(featurizer={}, classifier={}, thresholds_uri={})'.format(self.featurizer, self.classifier, self.thresholds_uri if self.thresholds_uri else self.rescaler.thresholds)
    #end def
#end class



__all__ = ['classification_report', 'find_best_thresholds', 'generate_pr_curves']

import logging
import os

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt

import numpy as np

from sklearn.metrics import accuracy_score
from sklearn.metrics import average_precision_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import precision_recall_fscore_support

from tabulate import tabulate

from uriutils import uri_open

logger = logging.getLogger(__name__)


def classification_report(Y_true, Y_proba, *, labels=None, target_names=None, thresholds=None, precision_thresholds=None, order='names'):
    Y_true, Y_proba = _make_label_indicator(Y_true, Y_proba)
    Y_true, Y_proba, target_names = _filter_labels(Y_true, Y_proba, labels=labels, target_names=target_names)

    n_classes = Y_true.shape[1]
    assert len(target_names) == n_classes

    if isinstance(thresholds, float): thresholds = np.full(n_classes, thresholds)
    if thresholds is not None and n_classes == 2: thresholds[1] = 1.0 - thresholds[0]
    if thresholds is not None:
        assert thresholds.shape[0] == n_classes
        assert ((thresholds <= 1).all() and (thresholds >= 0.0).all())
    #end if

    if isinstance(precision_thresholds, float): precision_thresholds = np.full(n_classes, precision_thresholds)
    if precision_thresholds is not None:
        assert precision_thresholds.shape[0] == n_classes
        assert ((precision_thresholds <= 1).all() and (precision_thresholds >= 0.0).all())
    #end if

    is_multiclass = all(np.isclose(Y_proba[i, :].sum(), 1.0) for i in range(Y_true.shape[0]))
    if is_multiclass:
        Y_predict = np.zeros(Y_proba.shape)
        for i in range(Y_true.shape[0]):
            j = np.argmax(Y_proba[i, :])
            Y_predict[i, j] = 1
        #end for
    else:
        Y_predict = (Y_proba >= 0.5)
    #end if

    table = []
    support_total, ap_score_total = 0.0, 0.0
    thresholds_best = np.zeros(n_classes)
    thresholds_minprec = np.zeros(n_classes)
    for i, name in enumerate(target_names):
        # Results using 0.5 as threshold
        # print(i, np.logical_and(Y_predict[:, i] == 1, Y_true[:, i] == 1).sum())
        if name == 'positive':
            print(Y_true[:, i].sum(), Y_predict[:, i].sum())
        p, r, f1, _ = precision_recall_fscore_support(Y_true[:, i], Y_predict[:, i], average='binary')  # Using default
        support = Y_true[:, i].sum()
        if support == 0: continue
        ap_score = average_precision_score(Y_true[:, i], Y_proba[:, i])
        row = [name, '{:d}'.format(int(support)), '{:.3f}'.format(ap_score), '{:.3f}/{:.3f}/{:.3f}'.format(p, r, f1)]
        support_total += support
        ap_score_total += ap_score

        # Results using given thresholds
        if thresholds is not None:
            p, r, f1, _ = precision_recall_fscore_support(Y_true[:, i], Y_proba[:, i] >= thresholds[i], average='binary')  # Using thresholds
            row.append('{:.3f}: {:.3f}/{:.3f}/{:.3f}'.format(thresholds[i], p, r, f1))
            # print(i, thresholds[i], (Y_proba[:, i] >= thresholds[i]).astype(int))
        #end if

        # Results using optimal threshold
        if n_classes == 2 and i == 1:
            thresholds_best[i] = 1.0 - thresholds_best[0]
            p, r, f1, _ = precision_recall_fscore_support(Y_true[:, i], Y_proba[:, i] >= thresholds_best[i], average='binary')
            row.append('{:.3f}: {:.3f}/{:.3f}/{:.3f}'.format(thresholds_best[i], p, r, f1))
        else:
            p, r, t = precision_recall_curve(Y_true[:, i], Y_proba[:, i])
            f1 = np.nan_to_num((2 * p * r) / (p + r + 1e-8))
            best_f1_i = np.argmax(f1)
            thresholds_best[i] = t[best_f1_i]
            # thresholds_best[i] = 0.5 if np.isclose(t[best_f1_i], 0.0) else t[best_f1_i]
            row.append('{:.3f}: {:.3f}/{:.3f}/{:.3f}'.format(thresholds_best[i], p[best_f1_i], r[best_f1_i], f1[best_f1_i]))
            # print(i, thresholds_best[i], (Y_proba[:, i] >= thresholds_best[i]).astype(int))
        #end if

        # Results using optimal threshold for precision > precision_threshold
        if precision_thresholds is not None:
            if n_classes == 2 and i == 1:
                thresholds_minprec[i] = 1.0 - thresholds_minprec[0]
                p, r, f1, _ = precision_recall_fscore_support(Y_true[:, i], Y_proba[:, i] >= thresholds_minprec[i], average='binary')
                row.append('{:.3f}: {:.3f}/{:.3f}/{:.3f}'.format(thresholds_minprec[i], p, r, f1))

            else:
                try:
                    best_f1_i = max(filter(lambda k: p[k] >= precision_thresholds[i], range(p.shape[0])), key=lambda k: f1[k])
                    if best_f1_i == p.shape[0] - 1 or f1[best_f1_i] == 0.0: raise ValueError()

                    thresholds_minprec[i] = t[best_f1_i]
                    # thresholds_minprec[i] = 0.5 if np.isclose(t[best_f1_i], 0.0) else t[best_f1_i]
                    row.append('{:.3f}: {:.3f}/{:.3f}/{:.3f}'.format(thresholds_minprec[i], p[best_f1_i], r[best_f1_i], f1[best_f1_i]))
                    # print(i, precision_thresholds[i], (Y_proba[:, i] >= precision_thresholds[i]).astype(int))

                except ValueError:
                    best_f1_i = np.argmax(f1)
                    logger.warning('Unable to find threshold for label "{}" where precision >= {}.'.format(target_names[i], precision_thresholds[i], t[best_f1_i]))
                    row.append('-')
                #end try
        #end if

        table.append(row)
    #end for

    if order == 'support':
        table.sort(key=lambda row: int(row[1]), reverse=True)

    headers = ['Label', 'Support', 'AP', 'Natural']
    if thresholds is not None: headers.append('File T')
    headers.append('Best T')
    if precision_thresholds is not None: headers.append('Min Prec T')

    if n_classes > 1:
        macro_averages = ['Macro average', '-', '{:.3f}'.format(average_precision_score(Y_true, Y_proba, average='macro')), '{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_predict, average='macro'))]
        if thresholds is not None: macro_averages.append('{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_proba >= thresholds, average='macro')))
        macro_averages.append('{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_proba >= thresholds_best, average='macro')))
        if precision_thresholds is not None: macro_averages.append('{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_proba >= thresholds_minprec, average='macro')))

        micro_averages = ['Micro average', '-', '{:.3f}'.format(average_precision_score(Y_true, Y_proba, average='micro')), '{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_predict, average='micro'))]
        if thresholds is not None: micro_averages.append('{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_proba >= thresholds, average='micro')))
        micro_averages.append('{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_proba >= thresholds_best, average='micro')))
        if precision_thresholds is not None: micro_averages.append('{:.3f}/{:.3f}/{:.3f}'.format(*precision_recall_fscore_support(Y_true, Y_proba >= thresholds_minprec, average='micro')))

        perfect_set = ['Perfect set', str(Y_true.shape[0]), '-', '{:.3f}'.format(accuracy_score(Y_true, Y_predict))]
        if thresholds is not None: perfect_set.append('{:.3f}'.format(accuracy_score(Y_true, Y_proba >= thresholds)))
        perfect_set.append('{:.3f}'.format(accuracy_score(Y_true, Y_proba >= thresholds_best)))
        if precision_thresholds is not None: perfect_set.append('{:.3f}'.format(accuracy_score(Y_true, Y_proba >= thresholds_minprec)))

        table += [perfect_set, macro_averages, micro_averages]
        table.insert(-3, ['-' * max(len(row[i]) for row in table + [headers]) for i in range(len(macro_averages))])
    #end if

    return tabulate(table, headers=headers, tablefmt='psql') + '\n{} labels, {} instances, {} instance-labels'.format(n_classes, Y_true.shape[0], int(support_total))
#end def


def find_best_thresholds(Y_true, Y_proba, *, labels=None, target_names=None, precision_thresholds=None):
    Y_true, Y_proba = _make_label_indicator(Y_true, Y_proba)
    Y_true, Y_proba, target_names = _filter_labels(Y_true, Y_proba, labels=labels, target_names=target_names)
    n_classes = Y_true.shape[1]

    if precision_thresholds is not None and isinstance(precision_thresholds, float): precision_thresholds = np.full(n_classes, precision_thresholds)

    assert Y_true.shape[0] == Y_proba.shape[0]
    assert Y_true.shape[1] == Y_proba.shape[1]
    assert len(target_names) == n_classes

    thresholds = np.zeros(n_classes)
    for i in range(n_classes):
        if n_classes == 2 and i == 1:
            thresholds[i] = 1.0 - thresholds[0]
            break
        #end if

        p, r, t = precision_recall_curve(Y_true[:, i], Y_proba[:, i])
        f1 = np.nan_to_num((2 * p * r) / (p + r + 1e-8))

        if precision_thresholds is None:  # use optimal threshold
            best_f1_i = np.argmax(f1)

        else:  # use optimal threshold for precision > precision_threshold
            try:
                best_f1_i = max(filter(lambda k: p[k] >= precision_thresholds[i], range(p.shape[0])), key=lambda k: f1[k])
                if best_f1_i == p.shape[0] - 1 or f1[best_f1_i] == 0.0: raise ValueError()

            except ValueError:
                best_f1_i = np.argmax(f1)
                logger.warning('Unable to find threshold for label "{}" where precision >= {}. Defaulting to best threshold of {}.'.format(target_names[i], precision_thresholds[i], t[best_f1_i]))
            #end try

        #end if

        thresholds[i] = t[best_f1_i]
    #end for

    return thresholds
#end def


def generate_pr_curves(Y_true, Y_proba, output_prefix, *, labels=None, target_names=None, thresholds=None, precision_thresholds=None):
    Y_true, Y_proba = _make_label_indicator(Y_true, Y_proba)
    Y_true, Y_proba, target_names = _filter_labels(Y_true, Y_proba, labels=labels, target_names=target_names)
    n_classes = Y_true.shape[1]

    if isinstance(thresholds, float): thresholds = np.full(n_classes, thresholds)
    if thresholds is not None and n_classes == 2: thresholds[1] = 1.0 - thresholds[0]
    assert thresholds is None or ((thresholds <= 1).all() and (thresholds >= 0.0).all())

    if isinstance(precision_thresholds, float): precision_thresholds = np.full(n_classes, precision_thresholds)
    assert precision_thresholds is None or ((precision_thresholds <= 1).all() and (precision_thresholds >= 0.0).all())

    thresholds_best = np.zeros(n_classes)
    thresholds_minprec = np.zeros(n_classes)
    for i, name in enumerate(target_names):
        if Y_true[:, i].sum() == 0 and Y_proba[:, i].sum() == 0: continue

        precision, recall, thresholds_ = precision_recall_curve(Y_true[:, i], Y_proba[:, i])
        f1 = np.nan_to_num((2 * precision * recall) / (precision + recall + 1e-8))
        ap_score = average_precision_score(Y_true[:, i], Y_proba[:, i])

        fig, ax = plt.subplots()
        ax.plot(recall, precision, label='Precision-Recall (AP={:.3f})'.format(ap_score))
        ax.set_ylabel('Precision')
        ax.set_xlabel('Recall')
        ax.set_ylim([0.0, 1.0])
        ax.set_xlim([0.0, 1.0])
        ax.set_title('Precision-Recall Curve for "{}"'.format(name))

        if thresholds is not None:  # Results using given thresholds
            p, r, f1_score, _ = precision_recall_fscore_support(Y_true[:, i], Y_proba[:, i] >= thresholds[i], average='binary')  # Using thresholds
            ax.plot([r], [p], marker='x', label='File T={:.3f}; F1={:.3f}'.format(thresholds[i], f1_score))
        #end if

        if n_classes == 2 and i == 1:  # Results using optimal threshold
            thresholds_best[i] = 1.0 - thresholds_best[0]
            p, r, f1_score, _ = precision_recall_fscore_support(Y_true[:, i], Y_proba[:, i] >= thresholds_best[i], average='binary')
        else:
            best_f1_i = np.argmax(f1)
            p, r, f1_score, thresholds_best[i] = precision[best_f1_i], recall[best_f1_i], f1[best_f1_i], thresholds_[best_f1_i]
        #end if
        ax.plot([r], [p], marker='x', label='Best T={:.3f}; F1={:.3f}'.format(thresholds_best[i], f1_score))

        if precision_thresholds is not None:  # Results using optimal threshold for precision > precision_threshold
            if n_classes == 2 and i == 1:
                thresholds_minprec[i] = 1.0 - thresholds_minprec[0]
                p, r, f1_score, _ = precision_recall_fscore_support(Y_true[:, i], Y_proba[:, i] >= thresholds_minprec[i], average='binary')

            else:
                try:
                    best_f1_i = max(filter(lambda k: precision[k] >= precision_thresholds[i], range(precision.shape[0])), key=lambda k: f1[k])
                    if best_f1_i == precision.shape[0] - 1 or f1[best_f1_i] == 0.0: raise ValueError()
                    thresholds_minprec[i] = thresholds_[best_f1_i]

                except ValueError:
                    best_f1_i = np.argmax(f1)
                    logger.warning('Unable to find threshold for label "{}" where precision >= {}.'.format(target_names[i], precision_thresholds[i], thresholds_[best_f1_i]))
                #end try
                p, r, f1_score, thresholds_minprec[i] = precision[best_f1_i], recall[best_f1_i], f1[best_f1_i], thresholds_[best_f1_i]
            #end if
            ax.plot([r], [p], marker='x', label='Minprec T={:.3f}; F1={:.3f}'.format(thresholds_minprec[i], f1_score))
        #end if

        ax.legend()
        plt.tight_layout()

        with uri_open(os.path.join(output_prefix, _sanitize_name(name) + '.txt'), 'w') as f:
            for p, r, t in zip(precision, recall, thresholds_):
                f.write('{} {} {}\n'.format(p, r, t))
        #end with

        with uri_open(os.path.join(output_prefix, _sanitize_name(name) + '.pdf'), 'wb') as f:
            fig.savefig(f, format='pdf')
            logger.info('Precision-Recall curve for "{}" saved to <{}>.'.format(name, f.name))
        #end with
    #end for
#end def


def _sanitize_name(name):
    return ''.join(c for c in name if c.isalnum() or c in ' ._-()+=&').rstrip()


def _make_label_indicator(Y_true, Y_proba):
    assert Y_true.shape[0] == Y_proba.shape[0]
    assert Y_proba.ndim == 2
    assert Y_proba.shape[1] > 1

    if Y_true.ndim == 1:
        Y_true_new = np.zeros(Y_proba.shape)
        Y_true_max = Y_true.max()

        if Y_true.shape[1] == 2:
            assert Y_true_max == 1
            for i in range(Y_true.shape[0]):
                Y_true_new[i, 0] = 1.0 - Y_true[i]
                Y_true_new[i, 1] = Y_true[i]
            #end for
        else:
            assert Y_true_max == Y_true.shape[1]
            for i in range(Y_true.shape[0]):
                Y_true_new[i, Y_true[i]] = 1
        #end if

        return Y_true_new, Y_proba
    #end if

    return Y_true, Y_proba
#end def


def _filter_labels(Y_true, Y_proba, labels=[], target_names=None):
    if target_names is None: target_names = list(range(Y_true.shape[1]))

    if labels:
        Y_true, Y_proba = Y_true[:, labels], Y_proba[:, labels]
        target_names = [target_names[j] for j in labels]
    #end if

    return Y_true, Y_proba, target_names
#end def


def main():
    Y_true = np.zeros((10, 3))
    Y_proba = np.random.rand(10, 3)

    Y_true[:4] = 1
    print(classification_report(Y_true, Y_proba, target_names=['ham', 'spam', 'bam'], precision_thresholds=0.75))
#end def


if __name__ == '__main__': main()



from argparse import ArgumentParser
from secrets import token_bytes, choice, token_hex
import string


def main():
    parser = ArgumentParser(description='Generate a secret key.')
    parser.add_argument('-l', '--length', type=int, default=32, help='Length of secret key in bytes.')
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-x', '--hex', action='store_true', help='Convert secret key to hexadecimal.')
    group.add_argument('-a', '--alphanum', action='store_true', help='Generate alphanumeric keys only.')
    A = parser.parse_args()

    if A.alphanum:
        alphabet = string.ascii_letters + string.digits
        print(''.join(choice(alphabet) for i in range(A.length)))
    elif A.hex: print(token_hex(A.length))
    else: print(token_bytes(A.length))
    #end if
#end def


if __name__ == '__main__': main()



from setuptools import setup, find_packages
import sys

if sys.version_info.major < 3:
    raise Exception('This is a Python 3 only package. Please upgrade.')
#end if

setup(
    name='ycml',
    version='0.1.9',
    author='yc sim',
    author_email='hello@yanchuan.sg',
    description='yc\'s collection of convenience code for developing ML applications.',
    long_description=open('README.md', 'r').read(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'Operating System :: POSIX',
    ],
    keywords=['machine learning', 'ml', 'natural language processing', 'nlp', 'utilities'],
    url='http://github.com/skylander86/ycml',
    license='Apache Software License 2.0',
    packages=find_packages('.'),
    install_requires=[line.strip() for line in open('requirements.txt', 'r') if not line.startswith('#') and line.strip()],
    include_package_data=True,
    zip_safe=False,
)



from argparse import ArgumentParser
import json
import logging
from operator import itemgetter
import os
from urllib.parse import urlparse

from flask import Flask
from flask import current_app, request, jsonify, abort

import numpy as np

try: import tensorflow as tf
except ImportError: tf = None

from uriutils import URIFileType

# We use ycml full paths here so it is easy to copy and paste this code.
from ycml.utils import get_settings, load_dictionary_from_file
from ycml.featclass import load_featclass
from ycml.http_daemon.decorators import check_api_token

logging.getLogger('botocore').setLevel(logging.WARNING)
logging.getLogger('boto3').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.WARNING)

logger = logging.getLogger('ycml.http_daemon')


def create_app(A, file_settings):
    log_level = get_settings(key='log_level', sources=('env', file_settings), default='INFO').upper()
    log_format = get_settings(key='log_format', sources=(A, 'env', file_settings), default='%(asctime)-15s [%(name)s-%(process)d] %(levelname)s: %(message)s')
    logging.basicConfig(format=log_format, level=logging.getLevelName(log_level))

    config_dict = dict(
        DEBUG=log_level in ['DEBUG'],
        CSRF_ENABLED=True,
        SECRET_KEY=os.urandom(24),
    )
    config_dict['JSONIFY_PRETTYPRINT_REGULAR'] = config_dict['DEBUG']

    api_token = get_settings(key='http_daemon_api_token', sources=('env', file_settings))
    if api_token is None:
        http_daemon_uri = get_settings(key='http_daemon_uri', sources=('env', file_settings), raise_on_missing=True)
        api_token = os.path.basename(urlparse(http_daemon_uri).path)
    #end if
    config_dict['api_token'] = api_token

    app = Flask('ycml.http_daemon')  # create our flask app!
    app.config.update(config_dict)

    with app.app_context():
        current_app.config['featclass'] = load_featclass(settings=file_settings, uri=get_settings(key='featclass_uri', sources=('env', file_settings)))
        if tf: current_app.config['tf_graph'] = tf.get_default_graph()
    #end with

    @app.route('/', methods=['GET', 'POST'])
    def index():
        return 'Toto, I\'ve a feeling we\'re not in Kansas anymore.', 404

    @app.route('/api/<api_token>/ping', methods=['GET'])
    @check_api_token
    def api_ping():
        return jsonify(dict(success=True))
    #end def

    @app.route('/api/<api_token>', methods=['POST'])
    @check_api_token
    def api():
        o = request.get_json(force=True, silent=True)
        if o is None: abort(400)

        try: remote_ip = request.headers.getlist('X-Forwarded-For')[0]
        except Exception: remote_ip = None

        instances = o.get('instances')
        if instances is None:
            instances = [o]

        if not instances:
            logger.debug('Request does not have any instances!')
            return jsonify(dict(reason='Request does not have any instances!')), 400
        #end if

        params = o.get('params', {})

        try:
            model = current_app.config['featclass']
            X = np.array(instances, dtype=np.object)

            if tf:
                with current_app.config['tf_graph'].as_default():
                    Y_proba, Y_predict = model.predict_and_proba(X, **params)
            else:
                Y_proba, Y_predict = model.predict_and_proba(X, **params)
            #end with

            limit = o.get('limit', False)
            predicted_only = o.get('predicted_only', False)
            probabilities = o.get('probabilities', False)

            if predicted_only:
                Y_proba = np.multiply(Y_proba, Y_predict)  # zeros out non predicted values

            if probabilities: Y, astype, epsilon = Y_proba, float, 0.0
            else: Y, astype, epsilon = Y_predict, bool, 0.0

            if hasattr(model.classifier, 'unbinarize_labels'): unbinarized = model.classifier.unbinarize_labels(Y, to_dict=True, astype=astype, epsilon=epsilon)
            else: unbinarized = [dict((j, astype(Y[i, j])) for j in range(Y.shape[1]) if Y[i, j] > epsilon) for i in range(Y.shape[0])]

            if limit: results = [dict(sorted(unbinarized[i].items(), key=itemgetter(1), reverse=True)[:limit]) for i in range(Y_proba.shape[0])]
            else: results = list(unbinarized)

            o['instances'] = '<{} instances>'.format(len(instances))
            logger.debug('Request={}; Response={}'.format(json.dumps(o), json.dumps(results)))

            return jsonify(results)

        except Exception:
            logging.exception('Exception while processing request from <{}>.'.format(remote_ip))
        #end try
    #end def

    return app
#end def


def gunicorn_app(environ, start_response):
    settings_uri = get_settings(key='settings_uri', sources=('env',), raise_on_missing=True)
    file_settings = load_dictionary_from_file(settings_uri)

    app = create_app({}, file_settings)

    return app(environ, start_response)
#end def


if __name__ == '__main__':
    parser = ArgumentParser(description='Starts the web daemon for URL classifier API.')
    parser.add_argument('settings_uri', type=URIFileType(), nargs='?', metavar='<settings_uri>', help='File containing daemon settings.')
    parser.add_argument('--settings', type=URIFileType(), metavar='<settings_uri>', help='File containing daemon settings.')
    parser.add_argument('--debug', action='store_true', default=False, help='Debug mode.')
    parser.add_argument('-p', '--port', type=int, default=None, metavar='p', help='Port to listen on.', dest='http_daemon_port')
    A = parser.parse_args()

    if A.settings: settings_uri = A.settings
    else: settings_uri = get_settings(key='settings_uri', sources=('env', A), raise_on_missing=True)

    file_settings = load_dictionary_from_file(settings_uri)

    http_daemon_port = get_settings(key='http_daemon_port', sources=('env', A, file_settings))
    if http_daemon_port is None:
        http_daemon_uri = get_settings(key='http_daemon_uri', sources=('env', file_settings), raise_on_missing=True)
        http_daemon_port = urlparse(http_daemon_uri).port
    #end if

    if http_daemon_port is None:
        http_daemon_port = 5000

    http_daemon_port = int(http_daemon_port)

    app = create_app(A, file_settings)
    app.run(debug=A.debug, host='0.0.0.0', use_reloader=False, port=http_daemon_port)
#end if
